"""
Influxdb reporter for Taurus stats

Copyright 2015 BlazeMeter Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import sys

import bzt
import time

from influxdb import InfluxDBClient

from bzt.engine import Reporter, Singletone
from bzt.modules.aggregator import DataPoint, KPISet, AggregatorListener, ResultsProvider
from bzt.utils import iteritems
from influxdb.exceptions import InfluxDBServerError, InfluxDBClientError


class InfluxdbStatusReporter(Reporter, AggregatorListener, Singletone):

    METRIC_COUNT = "count"
    METRIC_COUNT_ERROR = "countError"
    METRIC_HIT = "hit"
    METRIC_AVG = "avg"
    METRIC_RECEIVED_BYTES = "rb"
    METRIC_PCT_PREFIX = "pct"
    METRIC_MEAN_ACTIVE_THREADS = "meanAT"
    TAG_TRANSACTION = "transaction"
    TAG_STATUS = "status"
    TAG_APPLICATION = "application"
    TAG_RESPONSE_CODE = "responseCode"
    TAG_RESPONSE_MESSAGE = "responseMessage"
    TAG_OK = "ok"
    TAG_KO = "ko"
    TAG_ALL = "all"  # used for cumulated metrics

    """
    A reporter that send statistic status on test end to an influxdb datasource
    """

    def __init__(self):
        super(InfluxdbStatusReporter, self).__init__()
        self._common_tags = None
        self._client = None
        self.application = None
        self.measurement = None
        self._timestamp_serie = int(time.time())
        self.timeout = 30
        self.cumulative_results = None
        self.last_sec = None
        self.start_time = time.time()
        self.end_time = time.time()
        self.first_ts = sys.maxsize
        self.last_ts = 0
        self.kpi_buffer = []
        self.send_interval = 30
        self.send_data = True
        self._last_status_check = time.time()
        self._last_dispatch = 0

    def startup(self):
        self.start_time = time.time()

    def shutdown(self):
        self.end_time = time.time()

    def prepare(self):
        super(InfluxdbStatusReporter, self).prepare()

        self.send_data = self.parameters.get("send-data", self.send_data)
        self.application = self.parameters.get("application", self.settings.get("application", self.application))
        self.measurement = self.settings.get("measurement", None)
        host = self.settings.get("host", 'localhost')
        port = self.settings.get("port", 8086)
        db = self.settings.get("database", None)
        username = self.settings.get("username", None)
        password = self.settings.get("password", None)
        self._client = InfluxDBClient(host, port, username, password, db)

        # Build common tags
        self._common_tags = self.__compute_common_tags()

        # test connection
        if self.send_data:
            try:
                self._client.ping()  # check connectivity and auth
            except (InfluxDBServerError, InfluxDBClientError):
                self.log.error("Cannot reach influxdb, please check your settings.")
                raise

        if isinstance(self.engine.aggregator, ResultsProvider):
            self.engine.aggregator.add_listener(self)

    def post_process(self):
        super(InfluxdbStatusReporter, self).post_process()

        self.log.debug("Proceeding to %s KPI remaining report ...", len(self.kpi_buffer))

        self.log.info("Sending remaining KPI data to server...")
        self.__send_data(self.kpi_buffer)
        self.kpi_buffer = []

    # from AggregatorListener
    def aggregated_second(self, data):
        """
        Just store the latest info
        :type data: bzt.modules.aggregator.DataPoint
        """
        self.kpi_buffer.append(data)

    def check(self):
        """
        Send data if any in buffer at fixed interval
        """
        self.log.debug("KPI bulk buffer len: %s", len(self.kpi_buffer))
        if self._last_dispatch < (time.time() - self.send_interval):
            self._last_dispatch = time.time()
            if self.send_data and len(self.kpi_buffer):
                self.__send_data(self.kpi_buffer)
                self.kpi_buffer = []

        return super(InfluxdbStatusReporter, self).check()

    def __send_data(self, data):
        # depends on aggregator
        self.engine.aggregator.converter(data)
        series = self.__compute_data_points(data)

        if self.send_data:
            self.__send_series(series)

    def __compute_data_points(self, data):
        dps = []
        if data:
            # following data is received in the cumulative way
            # processing from last to first
            agg = data[-1]
            time_stamp = agg[DataPoint.TIMESTAMP]
            for label, kpi_set in iteritems(agg[DataPoint.CUMULATIVE]):
                if label == "":  # process total cumulated label
                    dps.extend(self.__compute_cumulative(kpi_set, time_stamp))

            # processing detailed metric for each label
            for dpoint in data:
                time_stamp = dpoint[DataPoint.TIMESTAMP]
                for label, kpi_set in iteritems(dpoint[DataPoint.CURRENT]):
                    self.log.debug('Computing timestamped metric for \'%s\' at %s', label, time_stamp)
                    dps.extend(self.__compute_current(label, kpi_set, time_stamp))

        return dps

    def __compute_cumulative(self, data, ts):
        series = []
        series.extend(self.__report_internal_metrics(data, '', ts))
        series.extend(self.__report_bytes(data, '', ts))
        # merge tags & common tags
        self.__merge_tags(series, self._common_tags)
        return series

    def __compute_current(self, label, data, ts):
        series = []

        series.extend(self.__report_detailed_labels(data, label, ts))

        # merge tags & common tags
        self.__merge_tags(series, self._common_tags)
        return series

    def __report_samples_count(self, kpi_set, sample_label=None):
        """
        reports samples count
        :kpi_set
        :returns metrics [dict, Any]
        """
        metrics = []
        transaction = self.TAG_ALL if sample_label == '' else sample_label
        # total count for the sample
        if kpi_set[KPISet.SAMPLE_COUNT]:
            metrics.append({
                "time": self._timestamp_serie,
                "measurement": self.measurement,
                "fields": {
                    self.METRIC_COUNT: kpi_set[KPISet.SAMPLE_COUNT],
                },
                "tags": {
                    self.TAG_TRANSACTION: transaction
                }
            })
            metrics.append({
                "time": self._timestamp_serie,
                "measurement": self.measurement,
                "fields": {
                    self.METRIC_HIT: kpi_set[KPISet.SAMPLE_COUNT],
                },
                "tags": {
                    self.TAG_TRANSACTION: transaction
                }
            })

        if kpi_set[KPISet.FAILURES]:
            metrics.append({
                "time": self._timestamp_serie,
                "measurement": self.measurement,
                "fields": {
                    self.METRIC_COUNT_ERROR: kpi_set[KPISet.FAILURES],
                },
                "tags": {
                    self.TAG_TRANSACTION: transaction
                }
            })

        return metrics

    def __report_detailed_labels(self, kpi_set, sample_label=None, ts=None):
        """
        reports detailed transactions
        kpi_set: DataPoint
        """
        metrics = []
        if sample_label != "":  # ignore cumulated dp
            """
            reports all transactions
            """
            metrics.append({
                "time": ts,
                "measurement": self.measurement,
                "fields": {
                    self.METRIC_COUNT: kpi_set['throughput'],
                    self.METRIC_AVG: kpi_set[KPISet.AVG_RESP_TIME]
                },
                "tags": {
                    self.TAG_TRANSACTION: sample_label,
                }
            })

            """
            reports successful transactions
            """
            successful_samples_count = kpi_set['succ']
            if successful_samples_count:
                metrics.append({
                    "time": ts,
                    "measurement": self.measurement,
                    "fields": {
                        self.METRIC_COUNT: successful_samples_count,
                    },
                    "tags": {
                        self.TAG_TRANSACTION: sample_label,
                        self.TAG_STATUS: self.TAG_OK
                    }
                })

            """
            reports failed transactions
            """
            failed_samples_count = kpi_set['fail']
            if failed_samples_count:
                for error in kpi_set['errors']:
                    metrics.append({
                        "time": ts,
                        "measurement": self.measurement,
                        "fields": {
                            self.METRIC_COUNT: error['cnt'],
                        },
                        "tags": {
                            self.TAG_TRANSACTION: sample_label,
                            self.TAG_STATUS: self.TAG_KO,
                            self.TAG_RESPONSE_CODE: error['rc'],
                            self.TAG_RESPONSE_MESSAGE: error['msg'] if 'msg' in error else None
                        }
                    })
            """
            reports percentiles
            """
            if self.parameters.get("percentiles", True):
                for key in sorted(kpi_set[KPISet.PERCENTILES].keys(), key=float):
                    metrics.append({
                        "time": ts,
                        "measurement": self.measurement,
                        "fields": {
                            self.METRIC_PCT_PREFIX + str(float(key)): kpi_set[KPISet.PERCENTILES][key],
                        },
                        "tags": {
                            self.TAG_TRANSACTION: sample_label
                        }
                    })
        return metrics

    def __report_internal_metrics(self, kpi_set, sample_label=None, ts=None):
        """
        reports internals (active users..)
        """
        metrics = []
        transaction = self.TAG_ALL if sample_label == '' else sample_label
        # average number of Virtual Users
        if kpi_set[KPISet.CONCURRENCY]:
            metrics.append({
                "time": ts,
                "measurement": self.measurement,
                "fields": {
                    self.METRIC_MEAN_ACTIVE_THREADS: kpi_set[KPISet.CONCURRENCY],
                },
                "tags": {
                    self.TAG_TRANSACTION: transaction
                }
            })
        return metrics

    def __report_bytes(self, kpi_set, sample_label=None, ts=None):
        """
        reports bytes received
        :returns: the metric formatted
        :rtype: [dict, Any]
        """
        metrics = []
        transaction = self.TAG_ALL if sample_label == '' else sample_label
        # total response size (header+body)
        if kpi_set[KPISet.BYTE_COUNT]:
            metrics.append({
                "time": ts,
                "measurement": self.measurement,
                "fields": {
                    self.METRIC_RECEIVED_BYTES: kpi_set[KPISet.BYTE_COUNT],
                },
                "tags": {
                    self.TAG_TRANSACTION: transaction
                }
            })
        return metrics

    def __compute_common_tags(self):
        common_tags = {
            self.TAG_TRANSACTION: self.TAG_ALL,
            self.TAG_STATUS: self.TAG_ALL,
        }
        if self.application is not None:
            common_tags[self.TAG_APPLICATION] = self.application
        return common_tags

    @staticmethod
    def __merge_tags(series, common_tags=None):
        """
        :param common_tags:
        :type series: Union[dict,str]
        """
        common_tags = common_tags or {}
        if series:
            for pv in series:
                if "tags" in pv:
                    merged_tags = {**common_tags, **pv["tags"]}
                    pv["tags"] = merged_tags
                else:
                    pv["tags"] = common_tags

    def __send_series(self, series=[]):
        """
        :type series: Union[dict,str]
        :return: dict
        """
        # by default epoch is set to 'ns'. time being computed as timestamp we override epoch to 's'.
        self._client.write_points(series, time_precision='s')



"""
Copyright 2018 BlazeMeter Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import ast
import re
from abc import abstractmethod

from bzt.utils import iteritems

from .ast_helpers import ast_call


class JMeterFunction(object):
    def __init__(self, arg_names, compiler):
        self.arg_names = arg_names
        self.compiler = compiler

    def to_python(self, arguments):
        """arguments -> (expression, stmts)"""
        args = dict(zip(self.arg_names, arguments))
        return self._compile(args)

    @abstractmethod
    def _compile(self, args):
        pass


class RandomFunction(JMeterFunction):
    def __init__(self, compiler):
        super(RandomFunction, self).__init__(["min", "max", "varname"], compiler)

    def _compile(self, args):
        if args.get("min") is None or args.get("max") is None:
            return None

        return ast.Call(
            func=ast.Attribute(
                value=ast.Name(id="apiritif", ctx=ast.Load()),
                attr='random_uniform',
                ctx=ast.Load(),
            ),
            args=[self.compiler.gen_expr(int(args["min"])), self.compiler.gen_expr(int(args["max"]))],
            keywords=[],
            starargs=None,
            kwargs=None
        )


class RandomStringFunction(JMeterFunction):
    def __init__(self, compiler):
        super(RandomStringFunction, self).__init__(["size", "chars", "varname"], compiler)

    def _compile(self, args):
        if args.get("size") is None:
            return None

        size = int(args.get("size"))
        arguments = [self.compiler.gen_expr(size)]
        if "chars" in args:
            arguments.append(self.compiler.gen_expr(args["chars"]))

        return ast.Call(
            func=ast.Attribute(
                value=ast.Name(id="apiritif", ctx=ast.Load()),
                attr='random_string',
                ctx=ast.Load(),
            ),
            args=arguments,
            keywords=[],
            starargs=None,
            kwargs=None
        )


class Base64DecodeFunction(JMeterFunction):
    def __init__(self, compiler):
        super(Base64DecodeFunction, self).__init__(["text"], compiler)

    def _compile(self, args):
        if args.get("text") is None:
            return None

        return ast.Call(
            func=ast.Attribute(
                value=ast.Name(id="apiritif", ctx=ast.Load()),
                attr='base64_decode',
                ctx=ast.Load(),
            ),
            args=[self.compiler.gen_expr(args["text"])],
            keywords=[],
            starargs=None,
            kwargs=None
        )


class Base64EncodeFunction(JMeterFunction):
    def __init__(self, compiler):
        super(Base64EncodeFunction, self).__init__(["text"], compiler)

    def _compile(self, args):
        if args.get("text") is None:
            return None

        return ast.Call(
            func=ast.Attribute(
                value=ast.Name(id="apiritif", ctx=ast.Load()),
                attr='base64_encode',
                ctx=ast.Load(),
            ),
            args=[self.compiler.gen_expr(args["text"])],
            keywords=[],
            starargs=None,
            kwargs=None
        )


class TimeFunction(JMeterFunction):
    def __init__(self, compiler):
        super(TimeFunction, self).__init__(["format", "varname"], compiler)

    def _compile(self, args):
        arguments = []
        if "format" in args:
            arguments.append(self.compiler.gen_expr(args["format"]))
        return ast.Call(
            func=ast.Attribute(
                value=ast.Name(id="apiritif", ctx=ast.Load()),
                attr='format_date',
                ctx=ast.Load(),
            ),
            args=arguments,
            keywords=[],
            starargs=None,
            kwargs=None
        )


class UrlEncodeFunction(JMeterFunction):
    def __init__(self, compiler):
        super(UrlEncodeFunction, self).__init__(["chars"], compiler)

    def _compile(self, args):
        if "chars" not in args:
            return None
        return ast.Call(
            func=ast.Attribute(
                value=ast.Name(id="apiritif", ctx=ast.Load()),
                attr='encode_url',
                ctx=ast.Load(),
            ),
            args=[self.compiler.gen_expr(args["chars"])],
            keywords=[],
            starargs=None,
            kwargs=None
        )


class UuidFunction(JMeterFunction):
    def __init__(self, compiler):
        super(UuidFunction, self).__init__([], compiler)

    def _compile(self, args):
        return ast.Call(
            func=ast.Attribute(
                value=ast.Name(id="apiritif", ctx=ast.Load()),
                attr='uuid',
                ctx=ast.Load(),
            ),
            args=[],
            keywords=[],
            starargs=None,
            kwargs=None
        )



class JMeterExprCompiler(object):
    def __init__(self, parent_log):
        self.log = parent_log.getChild(self.__class__.__name__)

    @staticmethod
    def gen_var_accessor(varname, ctx=None):
        if ctx is None:
            ctx = ast.Load()

        return ast.Subscript(
            value=ast.Name(id="self.vars", ctx=ast.Load()),
            slice=ast.Index(value=ast.Str(s=varname, kind="")),
            ctx=ctx
        )

    def gen_expr(self, value):
        if isinstance(value, bool):
            return ast.Name(id="True" if value else "False", ctx=ast.Load())
        elif isinstance(value, (int, float)):
            return ast.Num(n=value, kind="")
        elif isinstance(value, str):
            # if is has interpolation - break it into either a `"".format(args)` form or a Name node
            # otherwise - it's a string literal
            parts = re.split(r'(\$\{.*?\})', value)
            format_args = []
            for item in parts:
                if item:
                    if item.startswith("${") and item.endswith("}"):
                        value = value.replace(item, "{}")
                        compiled = self.translate_jmeter_expr(item[2:-1])
                        format_args.append(compiled)
            if format_args:
                if len(format_args) == 1 and value == "{}":
                    result = format_args[0]
                else:
                    result = ast_call(
                        func=ast.Attribute(
                            value=ast.Str(s=value, kind=""),
                            attr='format',
                            ctx=ast.Load()),
                        args=format_args)
            else:
                result = ast.Str(s=value, kind="")
            return result
        elif isinstance(value, type(None)):
            return ast.Name(id="None", ctx=ast.Load())
        elif isinstance(value, dict):
            items = sorted(list(iteritems(value)))
            return ast.Dict(keys=[self.gen_expr(k) for k, _ in items],
                            values=[self.gen_expr(v) for _, v in items])
        elif isinstance(value, list):
            return ast.List(elts=[self.gen_expr(val) for val in value], ctx=ast.Load())
        elif isinstance(value, tuple):
            return ast.Tuple(elts=[self.gen_expr(val) for val in value], ctx=ast.Load())
        elif isinstance(value, ast.AST):
            return value
        else:
            return value

    def translate_jmeter_expr(self, expr):
        """
        Translates JMeter expression into Apiritif-based Python expression.
        :type expr: str
        :return:
        """
        self.log.debug("Attempting to translate JMeter expression %r", expr)
        functions = {
            '__time': TimeFunction,
            '__Random': RandomFunction,
            '__RandomString': RandomStringFunction,
            '__base64Encode': Base64EncodeFunction,
            '__base64Decode': Base64DecodeFunction,
            '__urlencode': UrlEncodeFunction,
            '__UUID': UuidFunction,
        }
        regexp = r"(\w+)\((.*?)\)"
        args_re = r'(?<!\\),'
        match = re.match(regexp, expr)
        if match is None:  # doesn't look like JMeter func, translate as a var
            return self.gen_var_accessor(expr)

        varname, arguments = match.groups()

        if arguments is None:  # plain variable
            result = self.gen_var_accessor(varname)
        else:  # function call
            if not arguments:
                args = []
            else:
                # parse arguments: split by ',' but not '\,'
                args = [arg.strip() for arg in re.split(args_re, arguments)]

            if varname not in functions:  # unknown function
                return ast.Name(id=varname, ctx=ast.Load())

            self.log.debug("Translating function %s with arguments %s", varname, arguments)
            func = functions[varname](self)
            result = func.to_python(args)
            if result is None:
                result = ast.Name(id=varname, ctx=ast.Load())
        self.log.debug("Compile: %r -> %r", expr, result)
        return result




"""
Copyright 2017 BlazeMeter Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import logging
import os
import sys
import traceback
from optparse import OptionParser

from bzt import TaurusInternalException
from bzt.cli import CLI
from bzt.engine import Configuration
from bzt.utils import SoapUIScriptConverter, iteritems


class SoapUI2YAML(object):
    def __init__(self, options, file_name):
        self.log = logging.getLogger(self.__class__.__name__)
        self.options = options
        self.setup_logging()
        self.converter = None
        self.file_to_convert = file_name

    def setup_logging(self):
        CLI.setup_logging(self.options)
        if self.options.quiet:
            logging.disable(logging.WARNING)

    def process(self):
        output_format = Configuration.JSON if self.options.json else Configuration.YAML

        self.log.info('Loading SoapUI project %s', self.file_to_convert)
        self.file_to_convert = os.path.abspath(os.path.expanduser(self.file_to_convert))
        if not os.path.exists(self.file_to_convert):
            raise TaurusInternalException("File does not exist: %s" % self.file_to_convert)
        self.converter = SoapUIScriptConverter(self.log)
        try:
            converted_config = self.converter.convert_script(self.file_to_convert, self.options.test_case)
        except BaseException:
            self.log.error("Error while processing SoapUI project: %s", self.file_to_convert)
            raise

        self._cleanup_config(converted_config)

        exporter = Configuration.from_dict(converted_config)

        if self.options.file_name:
            file_name = self.options.file_name
        else:
            file_name = self.file_to_convert + "." + output_format.lower()

        exporter.dump(file_name, output_format)

        self.log.info("Done processing, result saved in %s", file_name)

    @staticmethod
    def _cleanup_config(config):
        for _, scenario in iteritems(config.get("scenarios")):
            if "test-case" in scenario:
                scenario.pop("test-case")
            if "test-suite" in scenario:
                scenario.pop("test-suite")


def process(parsed_options, args):
    tool = SoapUI2YAML(parsed_options, args[0])
    tool.process()


def main():
    usage = "Usage: soapui2yaml [input SoapUI file] [options]"
    parser = OptionParser(usage=usage, prog="soapui2yaml")
    parser.add_option('-t', '--test-case', action='store', default=None,
                      help="Extract only one test case given its name")
    parser.add_option('-v', '--verbose', action='store_true', default=False,
                      help="Prints all logging messages to console")
    parser.add_option('-o', '--out', dest="file_name",
                      help="Set output .yml file name, by default input file name + .yml is used")
    parser.add_option('-q', '--quiet', action='store_true', default=False, dest='quiet',
                      help="Do not display any log messages")
    parser.add_option('-j', '--json', action='store_true', default=False, dest='json',
                      help="Use JSON format")
    parser.add_option('-l', '--log', action='store', default=False, help="Log file location")
    parsed_options, args = parser.parse_args()
    if len(args) > 0:
        try:
            process(parsed_options, args)
        except BaseException as exc:
            logging.error("Exception: %s", exc)
            logging.debug("Exception: %s", traceback.format_exc())
            sys.exit(1)
        sys.exit(0)
    else:
        sys.stdout.write(usage + "\n")


if __name__ == "__main__":
    main()



"""
Module holds all stuff regarding JMeter tool usage

Copyright 2015 BlazeMeter Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import codecs
import copy
import csv
import fnmatch
import os
import re
import shutil
import socket
import tempfile
import time
import traceback
from collections import Counter, namedtuple
from distutils.version import LooseVersion
from itertools import dropwhile
from io import StringIO

from cssselect import GenericTranslator
from lxml import etree

from bzt import TaurusConfigError, ToolError, TaurusInternalException, TaurusNetworkError
from bzt.engine import Scenario, ScenarioExecutor
from bzt.engine import SETTINGS
from bzt.jmx import JMX, JMeterScenarioBuilder, LoadSettingsProcessor, try_convert
from bzt.modules.aggregator import ResultsReader, DataPoint, KPISet
from bzt.modules.console import ExecutorWidget
from bzt.modules.functional import FunctionalResultsReader, FunctionalSample
from bzt.requests_model import ResourceFilesCollector, has_variable_pattern, HierarchicRequestParser
from bzt.utils import iteritems, numeric_types, unicode_decode
from bzt.utils import get_full_path, EXE_SUFFIX, MirrorsManager, ExceptionalDownloader, get_uniq_name, is_windows
from bzt.utils import BetterDict, guess_csv_dialect, dehumanize_time, CALL_PROBLEMS
from bzt.utils import unzip, RequiredTool, JavaVM, shutdown_process, ProgressBarContext, TclLibrary, FileReader


def get_child_assertion(element):
    """
    Returns first failed assertion, or None

    :rtype lxml.etree.Element
    """
    for child in element.iterchildren():
        msg, name = parse_assertion(child)
        if msg:
            return msg, name

    return "", None


def parse_assertion(element, default=""):
    assertion = element.tag == "assertionResult"
    name = element.findtext("name")
    failure = element.findtext("failure") == "true"
    error = element.findtext("error") == "true"
    failure_message = element.findtext("failureMessage", default=default)
    if failure_message and failure_message.startswith("The operation lasted too long"):
        failure_message = "The operation lasted too long"

    wrong_message = "One or more sub-samples failed"

    failed_assertion = assertion and (failure or error) and (failure_message != wrong_message)

    if not failed_assertion:
        failure_message = ""

    return failure_message, name


class JMeterExecutor(ScenarioExecutor):
    """
    JMeter executor module

    :type modified_jmx: str
    :type jmeter_log: str
    :type properties_file: str
    :type sys_properties_file: str
    """
    UDP_PORT_NUMBER = None

    def __init__(self):
        super(JMeterExecutor, self).__init__()
        self.original_jmx = None
        self.modified_jmx = ""
        self.jmeter_log = None
        self.properties = BetterDict()
        self.properties_file = None
        self.sys_properties_file = None
        self.kpi_jtl = None
        self.log_jtl = None
        self.process = None
        self.end_time = None
        self.retcode = None
        self.distributed_servers = []
        self.management_port = None
        self.resource_files_collector = None
        self.tool = None

    def get_load(self):
        """
        Helper method to read load specification
        """
        load = self.get_specific_load()

        throughput = load.throughput
        concurrency = load.concurrency
        iterations = load.iterations
        steps = load.steps
        hold = load.hold
        ramp_up = load.ramp_up

        duration = 0

        if ramp_up is not None:
            ramp_up = try_convert(ramp_up, dehumanize_time, 0)
            duration += ramp_up

        if hold is not None:
            hold = try_convert(hold, dehumanize_time, 0)
            duration += hold

        msg = ''
        if not isinstance(concurrency, numeric_types + (type(None),)):
            msg += "\nNon-integer concurrency value [%s]: %s " % (type(concurrency).__name__, concurrency)
        if not isinstance(throughput, numeric_types + (type(None),)):
            msg += "\nNon-integer throughput value [%s]: %s " % (type(throughput).__name__, throughput)
        if not isinstance(steps, numeric_types + (type(None),)):
            msg += "\nNon-integer steps value [%s]: %s " % (type(steps).__name__, steps)
        if not isinstance(iterations, numeric_types + (type(None),)):
            msg += "\nNon-integer iterations value [%s]: %s " % (type(iterations).__name__, iterations)

        if msg:
            self.log.warning(msg)

        throughput = try_convert(throughput, float, default=0)
        concurrency = try_convert(concurrency, default=0)
        iterations = try_convert(iterations, default=0)
        steps = try_convert(steps, default=0)

        return self.LOAD_FMT(concurrency=concurrency, ramp_up=ramp_up, throughput=throughput, hold=hold,
                             iterations=iterations, duration=duration, steps=steps)

    def get_specific_load(self):
        """
        Helper method to read load specification
        """
        # throughput, concurrency, iterations, steps, hold, ramp_up
        raw_load = self.get_raw_load()

        hold = try_convert(raw_load.hold or 0, dehumanize_time)

        ramp_up = try_convert(raw_load.ramp_up, dehumanize_time)

        if not hold:
            duration = ramp_up
        elif not ramp_up:
            duration = hold
        elif isinstance(ramp_up, numeric_types) and isinstance(hold, numeric_types):
            duration = hold + ramp_up
        else:
            duration = 1  # dehumanize_time(<sum_of_props>) can be unpredictable so we use default there

        throughput = try_convert(raw_load.throughput, float)
        concurrency = try_convert(raw_load.concurrency)
        iterations = try_convert(raw_load.iterations)
        steps = try_convert(raw_load.steps)

        if not iterations:
            if duration:
                iterations = 0  # which means infinite
            else:
                iterations = 1

        return self.LOAD_FMT(concurrency=concurrency, ramp_up=ramp_up, throughput=throughput, hold=hold,
                             iterations=iterations, duration=duration, steps=steps)

    @staticmethod
    def _get_tool_version(jmx_file):
        jmx = JMX(jmx_file)
        selector = 'jmeterTestPlan'
        test_plan = jmx.get(selector)[0]
        ver = test_plan.get('jmeter')
        if isinstance(ver, str):
            space_index = ver.find(" ")
            # return original value if it doesn't contain the space
            # ver = "4.0" -> return "4.0" (NOT JMeter.VERSION)
            # ver = "2.13 r1665067" -> return "2.13"
            if space_index != -1:
                return ver[:space_index]
            dash_index = ver.find("-")
            if dash_index != -1:
                # not supported version like: 2.12-SNAPSHOT.20150128
                return JMeter.VERSION
            return ver
        return JMeter.VERSION

    def prepare(self):
        """
        Preparation for JMeter involves either getting existing JMX
        and modifying it, or generating new JMX from input data. Then,
        original JMX is modified to contain JTL writing classes with
        required settings and have workload as suggested by Provisioning

        :raise TaurusConfigError:
        """
        super(JMeterExecutor, self).prepare()
        self.jmeter_log = self.engine.create_artifact("jmeter", ".log")
        self._set_remote_port()
        self.distributed_servers = self.execution.get('distributed', self.distributed_servers)

        is_jmx_generated = False

        self.original_jmx = self.get_script_path()
        self.log.debug("Getting Jmeter version.")
        if os.getenv('TAURUS_JMETER_DISABLE_AUTO_DETECT', 'False') == 'True':
            self.log.debug("Autodetect version is disabled.")
            # autodetection NOT active -> use STABLE if auto,
            # xyz if STABLE|LATEST
            # STABLE otherwise
            config_version = self.settings.get("version", JMeter.VERSION, force_set=True)
            if config_version in [JMeter.VERSION, JMeter.VERSION_LATEST]:
                self.settings["version"] = config_version
            else:
                self.settings["version"] = JMeter.VERSION
        else:
            # autodetection IS active (original default behavior)
            self.log.debug("Autodetect version is enabled")
            if self.settings.get("version", JMeter.VERSION, force_set=True) == "auto":
                self.settings["version"] = self._get_tool_version(self.original_jmx)
        self.log.debug("JMeter version: %s", self.settings["version"])

        if not self.original_jmx:
            if self.get_scenario().get("requests"):
                self.original_jmx = self.__jmx_from_requests()
                is_jmx_generated = True
            else:
                raise TaurusConfigError("You must specify either a JMX file or list of requests to run JMeter")

        self.__set_jvm_properties()
        self.__set_system_properties()
        self.__set_jmeter_properties()

        self.install_required_tools()

        if self.engine.aggregator.is_functional:
            flags = {"connectTime": True}
            version = LooseVersion(self.tool.version)
            major = version.version[0]
            if major == 2:
                flags["bytes"] = True
            else:
                flags["sentBytes"] = True
            self.settings.merge({"xml-jtl-flags": flags})

        modified = self.__get_modified_jmx(self.original_jmx, is_jmx_generated)
        self.modified_jmx = self.__save_modified_jmx(modified, self.original_jmx, is_jmx_generated)

        # check for necessary plugins and install them if needed
        if self.settings.get("detect-plugins", True):
            self.tool.install_for_jmx(self.modified_jmx)

        self.stdout = open(self.engine.create_artifact("jmeter", ".out"), "w")
        self.stderr = open(self.engine.create_artifact("jmeter", ".err"), "w")

        if self.engine.is_functional_mode():
            self.reader = FuncJTLReader(self.log_jtl, self.engine, self.log)
            self.reader.is_distributed = len(self.distributed_servers) > 0
            self.reader.executor_label = self.label
            self.engine.aggregator.add_underling(self.reader)
        else:
            err_msg_separator = self.settings.get("error-message-separator")
            self.reader = JTLReader(self.kpi_jtl, self.log, self.log_jtl, err_msg_separator)
            self.reader.is_distributed = len(self.distributed_servers) > 0
            assert isinstance(self.reader, JTLReader)
            self.engine.aggregator.add_underling(self.reader)

    def __set_system_properties(self):
        sys_props = self.settings.get("system-properties")
        if sys_props:
            self.log.debug("Additional system properties %s", sys_props)
            self.properties.merge(sys_props)
            sys_props_file = self.engine.create_artifact("system", ".properties")
            JMeterExecutor.__write_props_to_file(sys_props_file, sys_props)
            self.sys_properties_file = sys_props_file

    def __set_jvm_properties(self):
        heap_size = self.settings.get("memory-xmx", None)
        if heap_size is not None:
            self.log.debug("Setting JVM heap size to %s", heap_size)
            self.env.add_java_param({"JVM_ARGS": "-Xmx%s" % heap_size})

    def __set_jmeter_properties(self):
        props = copy.deepcopy(self.settings.get("properties"))
        props_local = copy.deepcopy(self.get_scenario().get("properties"))
        if self.distributed_servers and self.settings.get("gui", False):
            props_local.merge({"remote_hosts": ",".join(self.distributed_servers)})
        props_local.update({"jmeterengine.nongui.port": self.management_port})
        props_local.update({"jmeterengine.nongui.maxport": self.management_port})
        props_local.update({"jmeter.save.saveservice.timestamp_format": "ms"})
        props_local.update({"sampleresult.default.encoding": "UTF-8"})
        props.merge(props_local)

        user_cp = [self.engine.artifacts_dir, get_full_path(self.original_jmx, step_up=1)]

        for _file in self.execution.get('files', []):
            full_path = get_full_path(_file)
            if os.path.isdir(full_path):
                user_cp.append(full_path)
            elif full_path.lower().endswith('.jar'):
                user_cp.append((get_full_path(_file, step_up=1)))

        if 'user.classpath' in props:
            user_cp.append(props['user.classpath'])

        props['user.classpath'] = os.pathsep.join(user_cp).replace(os.path.sep, "/")  # replace to avoid Windows issue

        self.log.debug("Additional properties: %s", props)
        self.properties.merge(props)
        self.properties_file = self.engine.create_artifact("jmeter-bzt", ".properties")
        JMeterExecutor.__write_props_to_file(self.properties_file, props)

    def startup(self):
        """
        Should start JMeter as fast as possible.
        """
        cmdline = [self.tool.tool_path, "-t", self.modified_jmx, "-j", self.jmeter_log, "-q", self.properties_file]
        if not self.settings.get("gui", False):
            cmdline += ["-n"]

        if self.distributed_servers:
            cmdline += ["-G", self.properties_file]

        if self.sys_properties_file:
            cmdline += ["-S", self.sys_properties_file]
        if self.distributed_servers and not self.settings.get("gui", False):
            cmdline += ['-R%s' % ','.join(self.distributed_servers)]

        # fix for JMeter 4.0 bug where jmeter.bat requires JMETER_HOME to be set
        if is_windows() and self.tool.version == "4.0" and not self.env.get("JMETER_HOME"):
            tool_dir = self.tool.tool_path
            if os.path.isfile(self.tool.tool_path):
                tool_dir = get_full_path(self.tool.tool_path, step_up=2)
            self.env.set({"JMETER_HOME": tool_dir})

        user_cmd = self.settings.get("cmdline")
        if user_cmd:
            cmdline += user_cmd.split(" ")

        self.process = self._execute(cmdline)

    def check(self):
        """
        Checks if JMeter is still running. Also checks if resulting JTL contains
        any data and throws exception otherwise.

        :return: bool
        :raise ToolError:
        """
        self.retcode = self.process.poll()
        if self.retcode is not None:
            if self.retcode != 0:
                raise ToolError("JMeter exited with non-zero code: %s" % self.retcode, self.get_error_diagnostics())

            return True
        return False

    def shutdown(self):
        """
        If JMeter is still running - let's stop it.
        """
        distr_multiplier = len(self.execution.get('distributed', [None]))  # 1 for regular, N of servers for distributed

        max_attempts = self.settings.get("shutdown-wait", 5) * distr_multiplier
        if self._process_stopped(1):
            return

        try:
            if not self.settings.get("gui", False):
                udp_sock = socket.socket(type=socket.SOCK_DGRAM)

                self.log.info("Sending Shutdown command to JMeter on port %d...", self.management_port)
                udp_sock.sendto(b"Shutdown", ("localhost", self.management_port))
                if self._process_stopped(max_attempts):
                    self.log.debug("JMeter stopped on Shutdown command")
                    return

                self.log.info("Sending StopTestNow command to JMeter on port %d...", self.management_port)
                udp_sock.sendto(b"StopTestNow", ("localhost", self.management_port))
                if self._process_stopped(max_attempts):
                    self.log.debug("JMeter stopped on StopTestNow command")
                    return
        finally:
            if not self._process_stopped(1):
                self.log.warning("JMeter process is still alive, killing it")
                shutdown_process(self.process, self.log)

        if self.start_time:
            self.end_time = time.time()
            self.log.debug("JMeter worked for %s seconds", self.end_time - self.start_time)

    def post_process(self):
        self.engine.existing_artifact(self.modified_jmx, True)
        super(JMeterExecutor, self).post_process()

    def has_results(self):
        if self.reader and self.reader.read_records:
            return True
        else:
            return False

    def _process_stopped(self, cycles):
        while cycles > 0:
            if not (self.process and self.process.poll() is None):
                return True
            cycles -= 1
            time.sleep(self.engine.check_interval)
        return False

    def _set_remote_port(self):
        """
        set management udp port
        :return:
        """

        if not JMeterExecutor.UDP_PORT_NUMBER:
            JMeterExecutor.UDP_PORT_NUMBER = self.settings.get("shutdown-port", 4445)
        else:
            JMeterExecutor.UDP_PORT_NUMBER += 1

        while not self.__port_is_free(JMeterExecutor.UDP_PORT_NUMBER):
            self.log.debug("Port %d is busy, trying next one", JMeterExecutor.UDP_PORT_NUMBER)
            if JMeterExecutor.UDP_PORT_NUMBER == 65535:
                TaurusInternalException("JMeter: no free ports for management interface")
            else:
                JMeterExecutor.UDP_PORT_NUMBER += 1

        self.management_port = JMeterExecutor.UDP_PORT_NUMBER
        self.log.debug("Using port %d for management", self.management_port)

    def __port_is_free(self, port_num):
        """
        :return: Bool
        """
        udp_sock = socket.socket(type=socket.SOCK_DGRAM)
        try:
            self.log.debug("Checking if port %d is free", port_num)
            udp_sock.bind(("localhost", port_num))
            udp_sock.close()
            self.log.debug("Port %d is free", port_num)
            return True
        except socket.error:
            self.log.debug("Port %d is busy", port_num)
            return False

    @staticmethod
    def __disable_listeners(jmx):
        """
        Set ResultCollector to disabled
        :param jmx: JMX
        :return:
        """
        sel = 'stringProp[name=filename]'
        xpath = GenericTranslator().css_to_xpath(sel)

        listeners = jmx.get('ResultCollector')
        for listener in listeners:
            file_setting = listener.xpath(xpath)
            if not file_setting or not file_setting[0].text:
                listener.set("enabled", "false")

    def __apply_test_mode(self, jmx):
        func_mode = self.engine.is_functional_mode()
        test_plan_selector = "jmeterTestPlan>hashTree>TestPlan"
        plans = jmx.get(test_plan_selector)
        if not plans:
            self.log.warning("No test plans, can't set test mode")
            return
        test_plan = plans[0]
        props = test_plan.xpath('boolProp[@name="TestPlan.functional_mode"]')
        if props:
            prop = props[0]
            prop.text = "true" if func_mode else "false"
        else:
            element = jmx._get_functional_mode_prop(func_mode)
            jmx.append(test_plan_selector, element)

    @staticmethod
    def __fill_empty_delimiters(jmx):
        delimiters = jmx.get("CSVDataSet>stringProp[name='delimiter']")
        for delimiter in delimiters:
            if not delimiter.text:
                delimiter.text = ','

    @staticmethod
    def __add_listener(lst, jmx):
        jmx.append(JMeterScenarioBuilder.TEST_PLAN_SEL, lst)
        jmx.append(JMeterScenarioBuilder.TEST_PLAN_SEL, etree.Element("hashTree"))

    def __add_result_listeners(self, jmx):
        if self.engine.is_functional_mode():
            self.__add_trace_writer(jmx)
        else:
            self.__add_result_writers(jmx)

    def __add_trace_writer(self, jmx):
        self.log_jtl = self.engine.create_artifact("trace", ".jtl")
        flags = self.settings.get('xml-jtl-flags')
        log_lst = jmx.new_xml_listener(self.log_jtl, True, flags)
        self.__add_listener(log_lst, jmx)

    def __add_result_writers(self, jmx):
        version = LooseVersion(self.tool.version)
        csv_flags = self.settings.get('csv-jtl-flags')
        if version < LooseVersion("2.13"):
            csv_flags['^connectTime'] = False

        self.kpi_jtl = self.engine.create_artifact("kpi", ".jtl")
        kpi_lst = jmx.new_kpi_listener(self.kpi_jtl, csv_flags)
        self.__add_listener(kpi_lst, jmx)

        verbose = self.engine.config.get(SETTINGS).get("verbose", False)
        jtl_log_level = self.execution.get('write-xml-jtl', "full" if verbose else 'error')

        xml_flags = self.settings.get('xml-jtl-flags')

        if jtl_log_level == 'error':
            self.log_jtl = self.engine.create_artifact("error", ".jtl")
            log_lst = jmx.new_xml_listener(self.log_jtl, False, xml_flags)
            self.__add_listener(log_lst, jmx)
        elif jtl_log_level == 'full':
            self.log_jtl = self.engine.create_artifact("trace", ".jtl")
            log_lst = jmx.new_xml_listener(self.log_jtl, True, xml_flags)
            self.__add_listener(log_lst, jmx)

    def __force_tran_parent_sample(self, jmx):
        scenario = self.get_scenario()
        if scenario.get("force-parent-sample", False):
            self.log.debug("Enforcing parent sample for transaction controller")
            jmx.set_text('TransactionController > boolProp[name="TransactionController.parent"]', 'true')

    def __get_modified_jmx(self, original, is_jmx_generated):
        """
        add two listeners to test plan:
            - to collect basic stats for KPIs
            - to collect detailed errors/trace info
        :return: path to artifact
        """
        jmx = JMX(original)

        if self.get_scenario().get("disable-listeners", not self.settings.get("gui", False)):
            JMeterExecutor.__disable_listeners(jmx)

        user_def_vars = self.get_scenario().get("variables")
        if user_def_vars:
            jmx.append(JMeterScenarioBuilder.TEST_PLAN_SEL, jmx.add_user_def_vars_elements(user_def_vars))
            jmx.append(JMeterScenarioBuilder.TEST_PLAN_SEL, etree.Element("hashTree"))

        headers = self.get_scenario().get_headers()
        if headers:
            jmx.append(JMeterScenarioBuilder.TEST_PLAN_SEL, JMX._get_header_mgr(headers))
            jmx.append(JMeterScenarioBuilder.TEST_PLAN_SEL, etree.Element("hashTree"))

        self.__apply_test_mode(jmx)
        self.__add_result_listeners(jmx)
        if not is_jmx_generated:
            self.__force_tran_parent_sample(jmx)
            version = LooseVersion(self.tool.version)
            if version >= LooseVersion("3.2"):
                self.__force_hc4_cookie_handler(jmx)
        self.__fill_empty_delimiters(jmx)

        self.__apply_modifications(jmx)
        LoadSettingsProcessor(self).modify(jmx, is_jmx_generated)

        return jmx

    def __force_hc4_cookie_handler(self, jmx):
        selector = "[testclass=CookieManager]"
        fix_counter = 0
        for node in jmx.get(selector):
            name = "CookieManager.implementation"
            if not node.get(name):
                val = "org.apache.jmeter.protocol.http.control.HC4CookieHandler"
                node.append(JMX._string_prop(name, val))
                fix_counter += 1
        if fix_counter:
            self.log.info('%s obsolete CookieManagers are found and fixed' % fix_counter)

    def __save_modified_jmx(self, jmx, original_jmx_path, is_jmx_generated):
        script_name, _ = os.path.splitext(os.path.basename(original_jmx_path))
        modified_script_name = "modified_" + script_name
        if is_jmx_generated:
            filename = self.engine.create_artifact(modified_script_name, ".jmx")
        else:
            script_dir = get_full_path(original_jmx_path, step_up=1)
            filename = get_uniq_name(script_dir, modified_script_name, ".jmx")
        jmx.save(filename)
        return filename

    def __jmx_from_requests(self):
        """
        Generate jmx file from requests
        :return:
        """
        filename = self.engine.create_artifact("requests", ".jmx")
        jmx = JMeterScenarioBuilder(self)
        jmx.save(filename)
        self.settings.merge(jmx.system_props)
        return filename

    @staticmethod
    def __write_props_to_file(file_path, params):
        """
        Write properties to file
        :param file_path:
        :param params:
        :return:
        """
        with open(file_path, 'w') as fds:
            for key, val in iteritems(params):
                fds.write("%s=%s\n" % (key, val))

    def get_widget(self):
        """
        Add progress widget to console screen sidebar

        :return:
        """
        if not self.widget:
            label = "%s" % self
            self.widget = ExecutorWidget(self, "JMeter: " + label.split('/')[1])
        return self.widget

    def __modify_resources_paths_in_jmx(self, jmx, file_list):
        """
        Modify resource files paths in jmx etree

        :param jmx: JMX
        :param file_list: list
        :return:
        """
        file_set = set(file_list)
        missed_files = []
        while file_set:
            filename = file_set.pop()
            file_path_elements = jmx.xpath('//stringProp[text()="%s"]' % filename)
            if not file_path_elements:
                missed_files.append(filename)
            for file_path_element in file_path_elements:
                basename = os.path.basename(filename)
                self.log.debug("Replacing JMX path %s with %s", file_path_element.text, basename)
                file_path_element.text = basename

        if missed_files:
            self.log.warning("Files not found in JMX: %s", missed_files)

    def _resolve_jmx_relpaths(self, resource_files_from_jmx):
        """
        Attempt to paths relative to JMX script itself.

        :param resource_files_from_jmx:
        :return:
        """
        resource_files = []
        script_basedir = os.path.dirname(get_full_path(self.original_jmx))
        for res_file in resource_files_from_jmx:
            if not os.path.exists(res_file):
                path_relative_to_jmx = os.path.join(script_basedir, res_file)
                if os.path.exists(path_relative_to_jmx):
                    self.log.info("Resolved resource file with path relative to JMX: %s", path_relative_to_jmx)
                    resource_files.append(path_relative_to_jmx)
                    continue
            resource_files.append(res_file)
        return resource_files

    def resource_files(self):
        """
        Get list of resource files, modify jmx file paths if necessary
        """
        # get all resource files from requests
        scenario = self.get_scenario()
        resource_files = self.res_files_from_scenario(scenario)

        self.original_jmx = self.get_script_path()
        if self.original_jmx:
            jmx = JMX(self.original_jmx)
            resource_files_from_jmx = JMeterExecutor.__get_resource_files_from_jmx(jmx)
            if resource_files_from_jmx:
                execution_files = self.execution.get('files', [], force_set=True)
                execution_files.extend(self._resolve_jmx_relpaths(resource_files_from_jmx))
                self.__modify_resources_paths_in_jmx(jmx.tree, resource_files_from_jmx)
                script_name, script_ext = os.path.splitext(os.path.basename(self.original_jmx))
                self.original_jmx = self.engine.create_artifact(script_name, script_ext)
                jmx.save(self.original_jmx)
                scenario[Scenario.SCRIPT] = self.original_jmx

        script = self.get_script_path()
        if script:
            resource_files.append(script)

        return resource_files

    @staticmethod
    def __get_resource_files_from_jmx(jmx):
        """
        Get list of resource files paths from jmx scenario
        :return: (file list)
        """
        resource_files = []
        exclude_elements = ['kg.apc.jmeter.jmxmon.JMXMonCollector', 'JSR223Listener',
                            'kg.apc.jmeter.vizualizers.CorrectedResultCollector',
                            'kg.apc.jmeter.reporters.FlexibleFileWriter', 'BSFListener',
                            'kg.apc.jmeter.dbmon.DbMonCollector', 'BeanShellListener', 'MailerResultCollector',
                            'kg.apc.jmeter.perfmon.PerfMonCollector', 'ResultCollector',
                            'kg.apc.jmeter.vizualizers.CompositeResultCollector',
                            'kg.apc.jmeter.reporters.LoadosophiaUploader']
        search_patterns = ["File.path", "filename", "BeanShellSampler.filename"]
        for pattern in search_patterns:
            resource_elements = jmx.tree.findall(".//stringProp[@name='%s']" % pattern)
            for resource_element in resource_elements:
                # check if none of parents are disabled
                parent = resource_element.getparent()
                parent_disabled = False
                while parent is not None:  # ?
                    if parent.get('enabled') == 'false' or parent.tag in exclude_elements:
                        parent_disabled = True
                        break
                    parent = parent.getparent()

                if resource_element.text and not parent_disabled and not has_variable_pattern(resource_element.text):
                    resource_files.append(resource_element.text)
        return resource_files

    def res_files_from_scenario(self, scenario):
        files = []
        for source in scenario.get_data_sources():
            files.append(source['path'])
        requests = scenario.get_requests(parser=HierarchicRequestParser)
        for req in requests:
            files.extend(self.res_files_from_request(req))
            self.resource_files_collector.clear_path_cache()
        return files

    def res_files_from_request(self, request):
        if self.resource_files_collector is None:
            self.resource_files_collector = ResourceFilesCollector(self)
        return self.resource_files_collector.visit(request)

    def __apply_modifications(self, jmx):
        """
        :type jmx: JMX
        """
        modifs = self.get_scenario().get("modifications")

        if 'disable' in modifs:
            self.__apply_enable_disable(modifs, 'disable', jmx)

        if 'enable' in modifs:
            self.__apply_enable_disable(modifs, 'enable', jmx)

        if 'set-prop' in modifs:
            items = modifs['set-prop']
            for path, text in iteritems(items):
                parts = path.split('>')
                if len(parts) < 2:
                    raise TaurusConfigError("JMeter: property selector must have at least 2 levels")
                sel_parts = ["[testname='%s']" % parts[0]]

                for add in parts[1:]:
                    sel_parts.append("[name='%s']" % add)
                selector = '>'.join(sel_parts)
                if not jmx.set_text(selector, text):
                    selector = '>'.join(sel_parts[:-1])
                    if jmx.get(selector):
                        jmx.append(selector, JMX._string_prop(parts[-1], text))
                    else:
                        self.log.warning("No elements matched for set-prop: %s", path)

    def __apply_enable_disable(self, modifs, action, jmx):
        items = modifs[action]
        if not isinstance(items, list):
            modifs[action] = [items]
            items = modifs[action]
        for name in items:
            candidates = jmx.get("[testname]")
            for candidate in candidates:
                if fnmatch.fnmatch(candidate.get('testname'), name):
                    jmx.set_enabled("[testname='%s']" % candidate.get('testname'),
                                    True if action == 'enable' else False)

    def install_required_tools(self):
        """
        check tools
        """
        self.tool = self._get_tool(JMeter, config=self.settings, props=self.properties)

        required_tools = [self._get_tool(JavaVM), self._get_tool(TclLibrary), self.tool]
        for tool in required_tools:
            if not tool.check_if_installed():
                tool.install()

        self.settings['path'] = self.tool.tool_path

    @staticmethod
    def __trim_jmeter_log(log_contents):
        lines = [line for line in log_contents.split("\n") if line]
        relevant_lines = list(dropwhile(lambda lin: "ERROR" not in lin, lines))
        if relevant_lines:
            return "\n".join(relevant_lines)
        else:
            return log_contents

    def get_error_diagnostics(self):
        diagnostics = []
        if self.stdout is not None:
            with codecs.open(self.stdout.name, encoding='utf-8') as fds:
                contents = fds.read().strip()
                if contents.strip():
                    diagnostics.append("JMeter STDOUT:\n" + contents)
        if self.stderr is not None:
            with codecs.open(self.stderr.name, encoding='utf-8') as fds:
                contents = fds.read().strip()
                if contents.strip():
                    diagnostics.append("JMeter STDERR:\n" + contents)
        if self.jmeter_log is not None and os.path.exists(self.jmeter_log):
            with codecs.open(self.jmeter_log, encoding='utf-8') as fds:
                log_contents = fds.read().strip()
                trimmed_log = self.__trim_jmeter_log(log_contents)
                if trimmed_log:
                    diagnostics.append("JMeter log:\n" + trimmed_log)
        return diagnostics


class JTLReader(ResultsReader):
    """
    Class to read KPI JTL
    :type errors_reader: JTLErrorsReader
    """

    def __init__(self, filename, parent_logger, errors_filename=None, err_msg_separator=None):
        super(JTLReader, self).__init__()
        self.is_distributed = False
        self.log = parent_logger.getChild(self.__class__.__name__)
        self.csvreader = IncrementalCSVReader(self.log, filename)
        self.read_records = 0
        if errors_filename:
            self.errors_reader = JTLErrorsReader(
                errors_filename, parent_logger, err_msg_separator, label_converter=self.get_mixed_label)
        else:
            self.errors_reader = None

    def set_aggregation(self, aggregation):
        super().set_aggregation(aggregation)
        if self.errors_reader:
            self.errors_reader.set_aggregation(aggregation)

    def _read(self, last_pass=False):
        """
        Generator method that returns next portion of data

        :type last_pass: bool
        """
        if self.errors_reader:
            self.errors_reader.read_file(last_pass)

        for row in self.csvreader.read(last_pass):
            label = unicode_decode(row["label"])
            if self.is_distributed:
                concur = int(row["grpThreads"])
                trname = row["Hostname"] + row["threadName"][:row["threadName"].rfind('-')]
            else:
                concur = int(row["allThreads"])
                trname = ""

            rtm = int(row["elapsed"]) / 1000.0
            ltc = int(row["Latency"]) / 1000.0
            if "Connect" in row:
                cnn = int(row["Connect"]) / 1000.0
            else:
                cnn = None

            rcd = row["responseCode"]
            if rcd.endswith('Exception'):
                rcd = rcd.split('.')[-1]

            if row["success"] != "true":
                error = row["responseMessage"] if row["responseMessage"] else 'unknown error'
            else:
                error = None

            byte_count = int(row.get("bytes", 0))

            tstmp = int(int(row["timeStamp"]) / 1000.0)
            self.read_records += 1
            yield tstmp, label, concur, rtm, cnn, ltc, rcd, error, trname, byte_count

    def _calculate_datapoints(self, final_pass=False):
        for point in super(JTLReader, self)._calculate_datapoints(final_pass):
            if self.errors_reader:
                err_details = self.errors_reader.get_data(point[DataPoint.TIMESTAMP])  # get only for labels we have
                for label in err_details:
                    if label in point[DataPoint.CURRENT]:
                        point[DataPoint.CURRENT][label][KPISet.ERRORS] = err_details[label]
                    else:
                        self.log.warning("Had error data but no KPISet %s: %s", label, err_details[label])

                for label, label_data in iteritems(point[DataPoint.CURRENT]):
                    if self._redundant_aggregation:
                        continue

                    if label in err_details:
                        pass
                    elif label_data[KPISet.ERRORS]:
                        self.log.warning("No details for errors of %s, dropped info: %s", label,
                                         label_data[KPISet.ERRORS])

            point[DataPoint.SOURCE_ID] = self.csvreader.file.name + "@" + str(id(self))
            yield point


class FuncJTLReader(FunctionalResultsReader):
    """
    Class to read trace.jtl
    :type filename: str
    :type parent_logger: logging.Logger
    """

    FILE_EXTRACTED_FIELDS = ["requestBody", "responseBody", "requestCookiesRaw"]

    def __init__(self, filename, engine, parent_logger):
        super(FuncJTLReader, self).__init__()
        self.executor_label = "JMeter"
        self.log = parent_logger.getChild(self.__class__.__name__)
        self.parser = etree.XMLPullParser(events=('end',), recover=True)
        self.engine = engine
        self.file = FileReader(filename=filename, parent_logger=self.log)
        self.failed_processing = False
        self.read_records = 0

    def read(self, last_pass=True):
        """
        Read the next part of the file
        """
        if self.failed_processing:
            return

        self.__read_next_chunk(last_pass)

        for _, elem in self.parser.read_events():
            if elem.getparent() is not None and elem.getparent().tag == 'testResults':
                sample = self._extract_sample(elem)
                self.read_records += 1

                elem.clear()
                while elem.getprevious() is not None:
                    del elem.getparent()[0]

                yield sample

    def __read_next_chunk(self, last_pass):
        while not self.failed_processing:
            read = self.file.get_bytes(size=1024 * 1024, decode=False)
            if not read or not read.strip():
                break

            try:
                self.parser.feed(read)
            except etree.XMLSyntaxError as exc:
                self.failed_processing = True
                self.log.debug("Error reading trace.jtl: %s", traceback.format_exc())
                self.log.warning("Failed to parse errors XML: %s", exc)

            if not last_pass:
                break

    def _write_sample_data(self, filename, contents):
        artifact = self.engine.create_artifact(filename, ".bin")
        with open(artifact, 'wb') as fds:
            fds.write(contents.encode('utf-8'))
        return artifact

    @staticmethod
    def _extract_sample_assertions(sample_elem):
        assertions = []
        for result in sample_elem.findall("assertionResult"):
            name = result.findtext("name")
            failed = result.findtext("failure") == "true" or result.findtext("error") == "true"
            error_message = ""
            if failed:
                error_message = result.findtext("failureMessage")
            assertions.append({"name": name, "isFailed": failed, "errorMessage": error_message})
        return assertions

    def _parse_http_headers(self, header_str):
        headers = {}
        for line in header_str.split("\n"):
            clean_line = line.strip()
            if ":" in clean_line:
                key, value = clean_line.split(":", 1)
                headers[key] = value
        return headers

    def _parse_http_cookies(self, cookie_str):
        cookies = {}
        clean_line = cookie_str.strip()
        if "; " in clean_line:
            for item in clean_line.split("; "):
                key, value = item.split("=", 1)
                cookies[key] = value
        return cookies

    def _extract_sample_extras(self, sample_elem):
        method = sample_elem.findtext("method")
        uri = sample_elem.findtext("java.net.URL")  # smells like Java automarshalling
        req_headers = sample_elem.findtext("requestHeader") or ""
        resp_headers = sample_elem.findtext("responseHeader") or ""
        req_cookies = sample_elem.findtext("cookies") or ""

        thread_id = sample_elem.get("tn")
        split = thread_id.split("-")
        thread_group = "-".join(split[:-1])

        sample_extras = {
            "responseCode": sample_elem.get("rc"),
            "responseMessage": sample_elem.get("rm"),
            "responseTime": int(sample_elem.get("t") or 0),
            "connectTime": int(sample_elem.get("ct") or 0),
            "latency": int(sample_elem.get("lt") or 0),
            "responseSize": int(sample_elem.get("by") or 0),
            "requestSize": int(sample_elem.get("sby") or 0),
            "requestMethod": method,
            "requestURI": uri,

            "threadId": thread_id,
            "threadGroup": thread_group,

            "assertions": self._extract_sample_assertions(sample_elem),
            "requestHeaders": self._parse_http_headers(req_headers),
            "responseHeaders": self._parse_http_headers(resp_headers),
            "requestCookies": self._parse_http_cookies(req_cookies),

            "requestBody": sample_elem.findtext("queryString") or "",
            "responseBody": sample_elem.findtext("responseData") or "",
            "requestCookiesRaw": req_cookies,
        }

        sample_extras["requestBodySize"] = len(sample_extras["requestBody"])
        sample_extras["responseBodySize"] = len(sample_extras["responseBody"])
        sample_extras["requestCookiesSize"] = len(sample_extras["requestCookiesRaw"])

        return sample_extras

    def __write_sample_data_to_artifacts(self, sample_extras):
        for file_field in self.FILE_EXTRACTED_FIELDS:
            contents = sample_extras.pop(file_field)
            if contents:
                filename = "sample-%s" % file_field
                artifact = self._write_sample_data(filename, contents)
                sample_extras[file_field] = artifact

    def _extract_sample(self, sample_elem):
        tstmp = int(float(sample_elem.get("ts")) / 1000.0)
        label = sample_elem.get("lb")
        duration = float(sample_elem.get("t")) / 1000.0
        success = sample_elem.get("s") == "true"

        if success:
            status = "PASSED"
            error_msg = ""
            error_trace = ""
        else:
            assertion = self.__get_failed_assertion(sample_elem)
            if assertion is not None:
                status = "FAILED"
                error_msg = assertion.find("failureMessage").text
                error_trace = ""
            else:
                status = "BROKEN"
                error_msg, error_trace = self.get_failure(sample_elem)

        if error_msg.startswith("The operation lasted too long"):
            error_msg = "The operation lasted too long"

        sample_extras = self._extract_sample_extras(sample_elem)
        self.__write_sample_data_to_artifacts(sample_extras)

        return FunctionalSample(test_case=label, test_suite=self.executor_label, status=status,
                                start_time=tstmp, duration=duration,
                                error_msg=error_msg, error_trace=error_trace,
                                extras=sample_extras, subsamples=[])

    def get_failure(self, element):
        """
        Returns failure message and a stack trace
        """
        r_code = element.get('rc')
        if r_code and r_code.startswith("2") and element.get('s') == "false":
            children = [elem for elem in element.iterchildren() if elem.tag == "httpSample"]
            for child in children:
                child_failure = self.get_failure(child)
                if child_failure:
                    return child_failure
        else:
            message = element.get('rm')
            response_data = element.find("responseData")
            if response_data is not None:
                trace = response_data.text
            else:
                trace = ""
            return message, trace

    @staticmethod
    def __get_failed_assertion(element):
        """
        Returns first failed assertion, or None
        :rtype lxml.etree.Element
        """
        assertions = [elem for elem in element.iterchildren() if elem.tag == "assertionResult"]
        for assertion in assertions:
            failed = assertion.find("failure")
            error = assertion.find("error")
            if failed.text == "true" or error.text == "true":
                return assertion
        return None


class IncrementalCSVReader(object):
    """
    JTL csv reader
    """

    def __init__(self, parent_logger, filename):
        self.buffer = StringIO()
        self.csv_reader = None
        self.log = parent_logger.getChild(self.__class__.__name__)
        self.indexes = {}
        self.partial_buffer = ""
        self.file = FileReader(filename=filename, parent_logger=self.log)
        self.read_speed = 1024 * 1024

    def read(self, last_pass=False):
        """
        read data from jtl
        yield csv row
        :type last_pass: bool
        """
        lines = self.file.get_lines(size=self.read_speed, last_pass=last_pass)

        lines_read = 0
        bytes_read = 0

        for line in lines:
            if not line.endswith("\n"):
                self.partial_buffer += line
                continue

            line = "%s%s" % (self.partial_buffer, line)
            self.partial_buffer = ""

            lines_read += 1
            bytes_read += len(line)

            if self.csv_reader is None:
                dialect = guess_csv_dialect(line, force_doublequote=True)
                self.csv_reader = csv.DictReader(self.buffer, [], dialect=dialect)
                self.csv_reader.fieldnames += line.strip().split(self.csv_reader.dialect.delimiter)
                self.log.debug("Analyzed header line: %s", self.csv_reader.fieldnames)
                continue

            self.buffer.write(line)

        if lines_read:
            self.log.debug("Read: %s lines / %s bytes (at speed %s)", lines_read, bytes_read, self.read_speed)
            self._tune_speed(bytes_read)

            self.buffer.seek(0)
            for row in self.csv_reader:
                yield row

            self.buffer.seek(0)
            self.buffer.truncate(0)

    def _tune_speed(self, bytes_read):
        if bytes_read >= self.read_speed:
            self.read_speed = min(8 * 1024 * 1024, self.read_speed * 2)
        elif bytes_read < self.read_speed / 2:
            self.read_speed = max(self.read_speed / 2, 1024 * 1024)


class JTLErrorsReader(object):
    """
    Reader for errors.jtl, which is in XML max-verbose format

    :type filename: str
    :type parent_logger: logging.Logger
    """
    url_xpath = GenericTranslator().css_to_xpath("java\\.net\\.URL")

    def __init__(self, filename, parent_logger, err_msg_separator=None, label_converter=None):
        # http://stackoverflow.com/questions/9809469/python-sax-to-lxml-for-80gb-xml/9814580#9814580
        super(JTLErrorsReader, self).__init__()
        self.log = parent_logger.getChild(self.__class__.__name__)
        self.parser = etree.XMLPullParser(events=('end',))
        self.file = FileReader(filename=filename, parent_logger=self.log)
        self.buffer = BetterDict()
        self.failed_processing = False
        self.err_msg_separator = err_msg_separator
        self.label_converter = label_converter
        self._redundant_aggregation = False

    def set_aggregation(self, aggregation):
        self._redundant_aggregation = aggregation

    def read_file(self, final_pass=False):
        """
        Read the next part of the file
        """
        start_size = os.path.getsize(self.file.name) if self.file.is_ready() else 0
        while not self.failed_processing:
            # we need to feed bytes, not a unicode string, into the parser
            read = self.file.get_bytes(size=1024 * 1024, decode=False)  # "Huge input lookup" error without capping :)
            if not read or not read.strip():
                break

            self.log.debug("Read bytes from error file: %s", len(read))

            try:
                self.parser.feed(read)
            except etree.XMLSyntaxError as exc:
                self.failed_processing = True
                self.log.debug("Error reading errors.jtl: %s", traceback.format_exc())
                self.log.warning("Failed to parse errors XML: %s", exc)

            for _, elem in self.parser.read_events():
                if elem.getparent() is not None and elem.getparent().tag == 'testResults':
                    self._parse_element(elem)
                    elem.clear()  # cleanup processed from the memory
                    while elem.getprevious() is not None:
                        del elem.getparent()[0]

            if not final_pass:
                break
            elif self.file.is_ready() and os.path.getsize(self.file.name) != start_size:
                self.log.debug("Error file size has changed %d=>%d while reading offset %d",
                               start_size, os.path.getsize(self.file.name), self.file.offset)
                break

    def _parse_element(self, elem):
        if elem.get('s'):
            result = elem.get('s')
        else:
            result = elem.xpath('success')[0].text
        if result == 'false':
            if elem.items():
                self._extract_standard(elem)
            else:
                self._extract_nonstandard(elem)

    def get_data(self, max_ts):
        """
        Get accumulated errors data up to specified timestamp
        """
        result = BetterDict()
        for t_stamp in sorted(self.buffer.keys()):
            if t_stamp >= max_ts + 1:
                break
            labels = self.buffer.pop(t_stamp)
            for label, label_data in iteritems(labels):
                res = result.get(label, [], force_set=True)
                for err_item in label_data:
                    KPISet.inc_list(res, ('msg', err_item['msg']), err_item)

        if result:
            self.log.debug("Got error info for %s, labels: %s", max_ts, result.keys())
        return result

    def _extract_standard(self, elem):
        t_stamp = int(elem.get("ts")) / 1000.0
        label = elem.get("lb")
        message = elem.get('rm')
        r_code = elem.get("rc")

        self._extract_common(elem, label, r_code, t_stamp, message)

    def _extract_common(self, elem, label, r_code, t_stamp, r_msg):
        f_msg, f_url, f_rc, f_tag, f_type = self.find_failure(elem, r_msg, r_code)

        if f_type == KPISet.ERRTYPE_SUBSAMPLE:
            url_counts = Counter({f_url: 1})
        else:
            urls = elem.xpath(self.url_xpath)
            if urls:
                url_counts = Counter({urls[0].text: 1})
            else:
                url_counts = Counter()

        err_item = KPISet.error_item_skel(f_msg, f_rc, 1, f_type, url_counts, f_tag)
        buf = self.buffer.get(t_stamp, force_set=True)
        if self._redundant_aggregation:
            label = self.label_converter(label=label, rc=r_code, msg=r_msg)
        KPISet.inc_list(buf.get(label, [], force_set=True), ("msg", f_msg), err_item)
        KPISet.inc_list(buf.get('', [], force_set=True), ("msg", f_msg), err_item)

    def _extract_nonstandard(self, elem):
        t_stamp = int(elem.findtext("timeStamp")) / 1000.0
        label = elem.findtext("label")
        message = elem.findtext("responseMessage")
        r_code = elem.findtext("responseCode")

        self._extract_common(elem, label, r_code, t_stamp, message)

    def find_failure(self, element, def_msg="", def_rc=None):
        """ returns (message, url, rc, tag, err_type) """
        rc = element.get("rc", default="")

        e_msg = ""
        url = None
        err_type = KPISet.ERRTYPE_ERROR

        a_msg, name = get_child_assertion(element)

        if not rc.startswith("2"):  # this sample is failed
            e_msg = element.get("rm", default="")
            url = element.xpath(self.url_xpath)
            url = url[0].text if url else element.get("lb")
        elif a_msg:
            err_type = KPISet.ERRTYPE_ASSERT
        elif element.get("s") == "false":  # has failed sub element, we should look deeper...
            for child in element.iterchildren():
                if child.tag in ("httpSample", "sample"):  # let's check sub samples..
                    e_msg, url, rc, name, err_type = self.find_failure(child)
                    if e_msg:
                        if err_type == KPISet.ERRTYPE_ERROR:  # replace subsample error
                            err_type = KPISet.ERRTYPE_SUBSAMPLE
                        break

        if not (err_type == KPISet.ERRTYPE_SUBSAMPLE) and self.err_msg_separator and (a_msg or e_msg):
            msg = self.err_msg_separator.join((a_msg, e_msg))
        elif e_msg:
            msg = e_msg
        else:
            msg = a_msg

        if not msg and def_msg:  # top level, empty result
            msg = def_msg
            if self.err_msg_separator:  # add appropriate separator to default msg
                msg = self.err_msg_separator + msg

        return msg, url, rc or def_rc, name, err_type


class XMLJTLReader(JTLErrorsReader, ResultsReader):
    def __init__(self, filename, parent_logger):
        super(XMLJTLReader, self).__init__(filename, parent_logger)
        self.items = []

    def _read(self, final_pass=False):
        self.read_file()
        while self.items:
            yield self.items.pop(0)

    def _parse_element(self, elem):
        tstmp = int(int(elem.get("ts")) / 1000.0)
        label = elem.get("lb")
        rtm = int(elem.get("t")) / 1000.0
        ltc = int(elem.get("lt")) / 1000.0 if "lt" in elem.attrib else 0
        cnn = int(elem.get("ct")) / 1000.0 if "ct" in elem.attrib else 0
        byte_count = int(elem.get("by")) if "by" in elem.attrib else 0
        concur = int(elem.get("na")) if "na" in elem.attrib else 0
        trname = ''

        rcd = elem.get("rc")
        message = self.find_failure(elem, def_msg=elem.get("rm"))[0]

        error = message if elem.get("s") == "false" else None
        self.items.append((tstmp, label, concur, rtm, cnn, ltc, rcd, error, trname, byte_count))


class JMeter(RequiredTool):
    """
    JMeter tool
    """
    PLUGINS_MANAGER_VERSION = "1.7"
    PLUGINS_MANAGER_LINK = 'https://search.maven.org/remotecontent?filepath=kg/apc/jmeter-plugins-manager/{version}/jmeter-plugins-manager-{version}.jar'
    COMMAND_RUNNER_VERSION = "2.2"
    COMMAND_RUNNER_LINK = 'https://search.maven.org/remotecontent?filepath=kg/apc/cmdrunner/{version}/cmdrunner-{version}.jar'
    VERSION = "5.4.3"
    VERSION_LATEST = "5.5"

    def __init__(self, config=None, props=None, **kwargs):
        settings = config or BetterDict()
        props = props or BetterDict()

        version = settings.get("version", JMeter.VERSION)
        jmeter_path = settings.get("path", "~/.bzt/jmeter-taurus/{version}/")
        jmeter_path = get_full_path(jmeter_path).format(version=version)

        download_link = settings.get("download-link", None)
        if download_link is not None:
            download_link = download_link.format(version=version)

        plugins_manager_settings = settings.get("plugins-manager")
        plugins_manager_version = plugins_manager_settings.get("version", JMeter.PLUGINS_MANAGER_VERSION)
        self.plugins_manager_link = plugins_manager_settings.get(
            "download-link", JMeter.PLUGINS_MANAGER_LINK).format(version=plugins_manager_version)

        command_runner_settings = settings.get("command-runner")
        command_runner_version = command_runner_settings.get("version", JMeter.COMMAND_RUNNER_VERSION)
        self.command_runner_link = command_runner_settings.get(
            "download-link", JMeter.COMMAND_RUNNER_LINK).format(version=command_runner_version)
        self.plugins = settings.get("plugins", [])

        super(JMeter, self).__init__(tool_path=jmeter_path, download_link=download_link, version=version, **kwargs)

        self.fix_jars = settings.get("fix-jars", True)
        self.mirror_manager = JMeterMirrorsManager(self.http_client, self.log, self.version)

        additional_jvm_props = self._get_jvm_props(props)
        for key in additional_jvm_props:
            self.env.add_java_param({"JVM_ARGS": "-D%s=%s" % (key, additional_jvm_props[key])})

    def _get_jvm_props(self, settings):
        props = {key: settings[key] for key in settings if key.startswith("jpgc.")}

        if self.http_client:
            props.update(self.http_client.get_proxy_props())

        return props

    def check_if_installed(self):
        end_str_l = os.path.join('bin', 'jmeter' + EXE_SUFFIX)
        end_str_s = os.path.join('bin', 'jmeter')

        if os.path.isfile(self.tool_path):
            if self.run_and_check():  # all ok, it's really tool path
                return True
            else:  # probably it's path to other tool)
                raise TaurusConfigError('JMeter: wrong tool path: %s' % self.tool_path)

        if os.path.isdir(self.tool_path):  # it's dir: fix tool path and install if needed
            self.tool_path = os.path.join(self.tool_path, end_str_l)
            return self.run_and_check()

        # similar to future jmeter directory
        if not (self.tool_path.endswith(end_str_l) or self.tool_path.endswith(end_str_s)):
            self.tool_path = os.path.join(self.tool_path, end_str_l)

        return False

    def run_and_check(self):
        self.log.debug("Trying JMeter..")
        jmlog = tempfile.NamedTemporaryFile(prefix="jmeter", suffix="log", delete=False)

        cmd_line = [self.tool_path, '-j', jmlog.name, '--version']
        try:
            out, err = self.call(cmd_line)
        except CALL_PROBLEMS as exc:
            msg = "JMeter check failed: %s" % exc
            if os.path.exists(self.tool_path):
                raise ToolError(msg)

            self.log.debug(msg)
            return False
        finally:
            jmlog.close()

        self.log.debug("JMeter check: %s / %s", out, err)
        return True

    def _pmgr_call(self, params):
        cmd = [self._pmgr_path()] + params
        return self.call(cmd)

    def install_for_jmx(self, jmx_file):
        if not os.path.isfile(jmx_file):
            self.log.warning("Script %s not found" % jmx_file)
            return

        params = ["install-for-jmx", jmx_file]

        try:
            out, err = self._pmgr_call(params)
        except CALL_PROBLEMS as exc:
            self.log.warning("Failed to detect plugins for %s: %s", jmx_file, exc)
            return

        self.log.debug("Try to detect plugins for %s\n%s\n%s", jmx_file, out, err)

        if err and "Wrong command: install-for-jmx" in err:  # old manager
            self.log.debug("pmgr can't discover jmx for plugins")

        if out and "Plugins manager will apply some modifications" in out:
            time.sleep(5)  # allow for modifications to complete

    def __install_jmeter(self, dest):
        if self.download_link:
            jmeter_dist = self._download(use_link=True)
        else:
            jmeter_dist = self._download()

        try:
            self.log.info("Unzipping %s to %s", jmeter_dist, dest)
            unzip(jmeter_dist, dest, 'apache-jmeter-%s' % self.version)
        finally:
            os.remove(jmeter_dist)

        # set exec permissions
        os.chmod(os.path.join(dest, 'bin', 'jmeter'), 0o755)
        os.chmod(os.path.join(dest, 'bin', 'jmeter' + EXE_SUFFIX), 0o755)

        if not self.check_if_installed():
            raise ToolError("Unable to run %s after installation!" % self.tool_name)

    def __download_additions(self, tools):
        downloader = ExceptionalDownloader(self.http_client)
        with ProgressBarContext() as pbar:
            for url, path in tools:
                _file = os.path.basename(url)
                self.log.info("Downloading %s from %s", _file, url)
                try:
                    downloader.get(url, path, reporthook=pbar.download_callback)
                except KeyboardInterrupt:
                    raise
                except BaseException as exc:
                    self.log.debug("Error details: %s", traceback.format_exc())
                    raise TaurusNetworkError("Error while downloading %s: %s" % (_file, exc))

    def __install_plugins_manager(self, pm_installer_path):
        installer = "org.jmeterplugins.repository.PluginManagerCMDInstaller"
        cmd_line = ["java", "-cp", pm_installer_path, installer]
        self.log.debug("Trying: %s", cmd_line)
        try:
            out, err = self.call(cmd_line)
        except CALL_PROBLEMS as exc:
            raise ToolError("Failed to install PluginsManager: %s" % exc)

        self.log.debug("Install PluginsManager: %s / %s", out, err)

    def __install_plugins(self):
        plugins_manager_cmd = self._pmgr_path()
        plugin_str = ",".join(self.plugins)
        self.log.info("Installing JMeter plugins: %s", plugin_str)
        cmd_line = [plugins_manager_cmd, 'install', plugin_str]
        self.log.debug("Trying: %s", cmd_line)

        try:
            out, err = self.call(cmd_line)
        except CALL_PROBLEMS as exc:
            raise ToolError("Failed to install plugins %s: %s" % (plugin_str, exc))

        self.log.debug("Install plugins: %s / %s", out, err)

        if out and "Plugins manager will apply some modifications" in out:
            time.sleep(5)  # allow for modifications to complete

    def _pmgr_path(self):
        dest = get_full_path(self.tool_path, step_up=2)
        return os.path.join(dest, 'bin', 'PluginsManagerCMD' + EXE_SUFFIX)

    def _get_jar_fixes(self, lib_dir):
        direct_install_tools = []

        if not self.fix_jars or LooseVersion(self.version) < LooseVersion('5.0.0'):
            return direct_install_tools

        # these jars should be replaced with newer version in order to fix some vulnerabilities
        # component name and download link in https://repo1.maven.org/maven2/
        affected_components = {
            # Needs to be <1.4.18 for old Jmeters https://stackoverflow.com/questions/30812293/com-thoughtworks-xstream-security-forbiddenclassexception
            "xstream": "com/thoughtworks/xstream/xstream/1.4.20/xstream-1.4.20.jar",
            "jackson-annotations": "com/fasterxml/jackson/core/jackson-annotations/2.14.1/jackson-annotations-2.14.1.jar",
            "jackson-core": "com/fasterxml/jackson/core/jackson-core/2.14.1/jackson-core-2.14.1.jar",
            "jackson-databind": "com/fasterxml/jackson/core/jackson-databind/2.14.1/jackson-databind-2.14.1.jar",
            "json-smart": "net/minidev/json-smart/2.4.8/json-smart-2.4.8.jar",
            "jsoup": "org/jsoup/jsoup/1.15.3/jsoup-1.15.3.jar",
            "snakeyaml": "org/yaml/snakeyaml/1.33/snakeyaml-1.33.jar",
            "okhttp": "com/squareup/okhttp3/okhttp/4.10.0/okhttp-4.10.0.jar",
            "commons-text": "org/apache/commons/commons-text/1.10.0/commons-text-1.10.0.jar",
            "xmlgraphics-commons": "org/apache/xmlgraphics/xmlgraphics-commons/2.8/xmlgraphics-commons-2.8.jar"}

        if LooseVersion(self.version) <= LooseVersion('5.4.2'):  # log4j must be fixed till jmeter 5.4.2
            affected_names = ["log4j-core", "log4j-api", "log4j-slf4j-impl", "log4j-1.2-api"]
            fixed_version = '2.19.0'
            maven_link = "org/apache/logging/log4j/{name}/{ver}/{name}-{ver}.jar"

            for name in affected_names:
                affected_components[name] = maven_link.format(name=name, ver=fixed_version)

        jar_files = [_file for _file in os.listdir(lib_dir) if _file.endswith(".jar")]
        for jar_file in jar_files:
            for comp_name in affected_components:
                if jar_file.startswith(comp_name):
                    download_link = "https://repo1.maven.org/maven2/" + affected_components[comp_name]
                    target_path = os.path.join(lib_dir, os.path.basename(download_link))
                    direct_install_tools.append([download_link, target_path])
                    break

        return direct_install_tools

    def _fix_jquery_in_jmeter(self, jmeter_dir):
        if not self.fix_jars or LooseVersion(self.version) < LooseVersion('5.0.0'):
            return

        # Fix CVE-2016-10707 in jquery
        jquery_src_dir = os.path.join(jmeter_dir, "jquery-dist-3.6.1")
        jquery_target_dir = os.path.join(jmeter_dir,
            "bin/report-template/sbadmin2-1.0.7/bower_components/jquery/")
        jquery_tar = os.path.join(jmeter_dir, "jquery-dist-3.6.1.tar.gz")
        self.__download_additions([["https://github.com/jquery/jquery-dist/archive/3.6.1.tar.gz",
                                    jquery_tar]])
        shutil.unpack_archive(jquery_tar, jmeter_dir)
        shutil.rmtree(jquery_target_dir, ignore_errors=True)
        shutil.move(jquery_src_dir, jquery_target_dir)
        os.remove(jquery_tar)

    def install(self):
        jmeter_dir = get_full_path(self.tool_path, step_up=2)
        lib_dir = os.path.join(jmeter_dir, 'lib')
        self.log.info("Will install %s into %s", self.tool_name, jmeter_dir)
        plugins_manager_name = os.path.basename(self.plugins_manager_link)
        command_runner_name = os.path.basename(self.command_runner_link)
        pm_installer_path = os.path.join(lib_dir, 'ext', plugins_manager_name)
        command_runner_path = os.path.join(lib_dir, command_runner_name)

        self.__install_jmeter(jmeter_dir)

        direct_install_tools = [[self.plugins_manager_link, pm_installer_path],
                                [self.command_runner_link, command_runner_path]]
        self.__download_additions(direct_install_tools)
        self.__install_plugins_manager(pm_installer_path)
        self.__install_plugins()
        # Apply JAR vulnerability fixes
        self.__download_additions(self._get_jar_fixes(lib_dir))
        self._fix_jquery_in_jmeter(jmeter_dir)

        cleaner = JarCleaner(self.log)
        cleaner.clean(lib_dir)

    def ctg_plugin_installed(self):
        """
        Simple check if ConcurrentThreadGroup is available
        :return:
        """
        ext_dir = os.path.join(get_full_path(self.tool_path, step_up=2), 'lib', 'ext')
        if os.path.isdir(ext_dir):
            list_of_jars = [file_name for file_name in os.listdir(ext_dir) if file_name.endswith('.jar')]
            if any([file_name.startswith('jmeter-plugins-casutg') for file_name in list_of_jars]):
                return True

        return False


class JarCleaner(object):
    def __init__(self, parent_logger):
        self.log = parent_logger.getChild(self.__class__.__name__)

    @staticmethod
    def __extract_version(jar):
        version_str = jar.split('-')[-1]
        return version_str.replace('.jar', '')

    def clean(self, path):
        """
        Remove old jars
        :param path: str
        """
        self.log.debug("Removing old jars from %s", path)
        jarlib = namedtuple("jarlib", ("file_name", "lib_name", "version"))
        jars = [fname for fname in os.listdir(path) if '-' in fname and os.path.isfile(os.path.join(path, fname))]
        jar_libs = [jarlib(file_name=jar,
                           lib_name='-'.join(jar.split('-')[:-1]),
                           version=JarCleaner.__extract_version(jar))
                    for jar in jars]

        duplicated_libraries = set()
        for jar_lib_obj in jar_libs:
            similar_packages = [lib for lib in jar_libs if lib.lib_name == jar_lib_obj.lib_name]
            if len(similar_packages) > 1:
                right_version = max(similar_packages, key=lambda l: LooseVersion(l.version))
                similar_packages.remove(right_version)
                duplicated_libraries.update(similar_packages)

        for old_lib in duplicated_libraries:
            os.remove(os.path.join(path, old_lib.file_name))
            self.log.debug("Old jar removed %s", old_lib.file_name)


class JMeterMirrorsManager(MirrorsManager):
    MIRRORS_SOURCE = "https://jmeter.apache.org/download_jmeter.cgi"
    DOWNLOAD_LINK = "https://archive.apache.org/dist/jmeter/binaries/apache-jmeter-{version}.zip"

    def __init__(self, http_client, parent_logger, jmeter_version):
        self.jmeter_version = jmeter_version
        super(JMeterMirrorsManager, self).__init__(http_client, self.MIRRORS_SOURCE, parent_logger)

    def _parse_mirrors(self):
        links = []
        if self.page_source is not None:
            self.log.debug('Parsing mirrors...')
            select_search_pattern = re.compile(r'<select name="Preferred">.*?</select>', re.MULTILINE | re.DOTALL)
            option_search_pattern = re.compile(r'<option value=".*?">')
            select_element = select_search_pattern.findall(self.page_source)
            if select_element:
                option_elements = option_search_pattern.findall(select_element[0])
                link_tail = "/jmeter/binaries/apache-jmeter-{version}.zip".format(version=self.jmeter_version)
                links = [link.strip('<option value="').strip('">') + link_tail for link in option_elements]
        links.append(self.DOWNLOAD_LINK.format(version=self.jmeter_version))
        self.log.debug('Total mirrors: %d', len(links))
        # place HTTPS links first, preserving the order of HTTP links
        sorted_links = sorted(links, key=lambda l: l.startswith("https"), reverse=True)
        return sorted_links



"""
Implementations for `Provisioning` classes

Copyright 2015 BlazeMeter Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import datetime
import sys
import time
import traceback

from bzt import ToolError
from bzt.engine import Provisioning
from bzt.utils import dehumanize_time, reraise


class Local(Provisioning):
    """
    Local provisioning means we start all the tools locally
    """

    def __init__(self):
        super(Local, self).__init__()
        self.extend_configs = True
        self.start_time = None
        self.available_slots = None
        self.finished_modules = []
        self.started_modules = []

    def _get_start_shift(self, shift):
        if not shift:
            return 0

        time_formats = ['%Y-%m-%d %H:%M:%S',
                        '%Y-%m-%d %H:%M',
                        '%H:%M:%S',
                        '%H:%M']

        for time_format in time_formats:
            try:
                date = datetime.datetime.strptime(shift, time_format)
            except ValueError:
                continue
            except TypeError:
                self.log.warning('Start time must be string type ("%s"), ignored "%s"', time_format[0], shift)
                break
            today = datetime.date.today()
            if today > date.date():
                date = datetime.datetime(today.year, today.month, today.day, date.hour, date.minute, date.second)
            return time.mktime(date.timetuple()) - self.start_time
        else:
            self.log.warning('Unrecognized time format: %s ("%s" required), ignored', shift, time_formats[0])

        return 0

    def prepare(self):
        super(Local, self).prepare()
        for executor in self.executors:
            self.log.debug("Preparing executor: %s", executor)
            executor.prepare()

    def startup(self):
        self.start_time = time.time()

        if self.settings.get("sequential", False):
            self.available_slots = 1
        else:
            self.available_slots = self.settings.get("capacity", None)
            if not self.available_slots:
                self.available_slots = sys.maxsize  # no limit

        for executor in self.executors:
            start_at = executor.execution.get('start-at', 0)
            start_shift = self._get_start_shift(start_at)
            delay = dehumanize_time(executor.execution.get('delay', 0))
            executor.delay = delay + start_shift
            msg = "Delay setup for %s: %s(start-at) + %s(delay) = %s"
            self.log.debug(msg, executor, start_shift, delay, executor.delay)

    def _start_modules(self):
        if self.available_slots:
            non_started_executors = [e for e in self.executors if e not in self.started_modules]
            for executor in non_started_executors:
                self.engine.logging_level_up()
                if time.time() >= self.start_time + executor.delay:
                    executor.startup()
                    self.started_modules.append(executor)
                    self.available_slots -= 1
                    msg = "Starting execution: %s, rest of available slots: %s"
                    self.log.debug(msg, executor, self.available_slots)
                    if not self.available_slots:
                        break

                self.engine.logging_level_down()

    def check(self):
        """
        Check executors for finish. Return True if all of them has finished.
        """
        finished = True

        self._start_modules()
        for executor in self.executors:
            if executor in self.finished_modules:
                continue

            if executor not in self.started_modules:
                finished = False
                continue

            if executor.check():
                self.finished_modules.append(executor)
                self.available_slots += 1
                self.log.debug("%s finished", executor)
            else:
                finished = False

        return finished

    def shutdown(self):
        """
        Call shutdown on executors
        """
        exc_info = exc_value = None
        for executor in self.started_modules:
            self.log.debug("Shutdown %s", executor)
            try:
                executor.shutdown()
            except BaseException as exc:
                msg = "Exception in shutdown of %s: %s %s"
                self.log.debug(msg, executor.__class__.__name__, exc, traceback.format_exc())
                if not exc_info:
                    exc_info = sys.exc_info()
                if not exc_value:
                    exc_value = exc
        if exc_info:
            reraise(exc_info, exc_value)

    def post_process(self):
        """
        Post-process executors
        """
        exc_info = exc_value = None
        for executor in self.executors:
            self.log.debug("Post-process %s", executor)
            try:
                executor.post_process()
                if executor in self.started_modules and not executor.has_results():
                    msg = "Empty results, most likely %s (%s) failed. " \
                          "Actual reason for this can be found in logs under %s"
                    message = msg % (executor.label, executor.__class__.__name__, self.engine.artifacts_dir)
                    diagnostics = executor.get_error_diagnostics()
                    raise ToolError(message, diagnostics)
            except BaseException as exc:
                msg = "Exception in post_process of %s: %s %s"
                self.log.debug(msg, executor.__class__.__name__, exc, traceback.format_exc())
                if not exc_info:
                    exc_info = sys.exc_info()
                if not exc_value:
                    exc_value = exc
        if exc_info:
            reraise(exc_info, exc_value)



import os
import re
import time
from glob import glob

from bzt import TaurusInternalException
from bzt.engine import ScenarioExecutor
from bzt.modules.ab import TSVDataReader
from bzt.modules.aggregator import AggregatorListener, ConsolidatingAggregator, DataPoint
from bzt.modules.gatling import DataLogReader as GatlingLogReader
from bzt.modules.jmeter import JTLReader, XMLJTLReader
from bzt.utils import dehumanize_time


class ExternalResultsLoader(ScenarioExecutor, AggregatorListener):
    """
    :type reader: bzt.modules.aggregator.ResultsReader
    """
    AB_HEADER = "starttime\tseconds\tctime\tdtime\tttime\twait"
    PBENCH_FORMAT = re.compile("^[0-9]+\.[0-9]{3}\t[^\t]*\t([0-9]+\t){9}[0-9]+$")

    def __init__(self):
        super(ExternalResultsLoader, self).__init__()
        self._result_timeout = 1
        self._file_exists_wait = 1
        self._data_file_pattern = None
        self.data_file = None
        self.errors_file = None
        self.reader = None
        self._last_update_ts = -1
        self._last_datapoint_ts = -1
        self._prev_datapoint_ts = -1
        self._file_check_ts = time.time()

    def prepare(self):
        super(ExternalResultsLoader, self).prepare()
        self._read_options()
        assert self._data_file_pattern or self.data_file, "Option is required: data-file or data-file-pattern"
        self.label = self.data_file
        if self.errors_file:
            self.errors_file = self.engine.find_file(self.errors_file)

        str_wait = self.execution.get("wait-for-file", self.settings.get("wait-for-file", self._file_exists_wait))
        self._file_exists_wait = dehumanize_time(str_wait)

        def_timout = self.engine.check_interval * 10
        str_to = self.execution.get("results-timeout", self.settings.get("results-timeout", def_timout))
        self._result_timeout = dehumanize_time(str_to)

        self._file_check_ts = time.time()
        self._try_make_reader()

    def _read_options(self):
        # read from scenario
        if 'scenario' in self.execution:
            scenario = self.get_scenario()
            self.data_file = scenario.get('data-file', self.data_file)
            self.errors_file = scenario.get('errors-file', self.errors_file)
            self._data_file_pattern = scenario.get("data-file-pattern", self._data_file_pattern)

        # execution level overrides scenario level
        self.data_file = self.execution.get("data-file", self.data_file)
        self.errors_file = self.execution.get("errors-file", self.errors_file)
        self._data_file_pattern = self.execution.get("data-file-pattern", self._data_file_pattern)

    def _try_make_reader(self):
        if self.reader:
            return

        if not self.data_file:
            files = glob(self._data_file_pattern)
            if not files:
                return
            files.sort()
            self.log.debug("Files found by pattern: %s", files)
            if not os.path.getsize(files[-1]):
                return
            self.data_file = files[-1]

        self.data_file = self.engine.find_file(self.data_file)
        if not os.path.exists(self.data_file):
            if time.time() - self._file_check_ts < self._file_exists_wait:
                self.log.debug("File not exists yet: %s", self.data_file)
                return
            else:
                msg = "File has not appeared within %ss: %s" % (self._file_exists_wait, self.data_file)
                raise TaurusInternalException(msg)

        self.log.info("Will load external results from file: %s", self.data_file)
        self.label = self.data_file

        self.reader = self._get_reader()
        if isinstance(self.engine.aggregator, ConsolidatingAggregator):
            self.engine.aggregator.add_underling(self.reader)
            self.engine.aggregator.add_listener(self)

    def _get_reader(self):
        with open(self.data_file) as fhd:
            header = fhd.readline(2048).strip()  # just header chunk of file

        if header.startswith(self.AB_HEADER):
            reader = TSVDataReader(self.data_file, self.log)
            reader.url_label = "N/A"
            return reader
        elif header.startswith("<?xml"):
            return XMLJTLReader(self.data_file, self.log)
        elif header.startswith("RUN\t") or "\tRUN\t" in header:
            return GatlingLogReader(self.data_file, self.log, None)
        elif "timestamp" in header.lower() and "elapsed" in header.lower():
            return JTLReader(self.data_file, self.log, self.errors_file)
        else:
            self.log.info("Header line was: %s", header)
            raise TaurusInternalException("Unable to detect results format for: %s" % self.data_file)

    def aggregated_second(self, data):
        self._last_datapoint_ts = data[DataPoint.TIMESTAMP]
        self._last_update_ts = time.time()
        # self.log.debug("Notified of datapoint %s", self._last_datapoint_ts)

    def startup(self):
        super(ExternalResultsLoader, self).startup()
        self._try_make_reader()

    def check(self):
        self._try_make_reader()
        ts_not_changed = self._last_datapoint_ts == self._prev_datapoint_ts
        no_new_results = time.time() - self._last_update_ts > self._result_timeout
        has_read_some = self._last_datapoint_ts > 0 or bool(self.reader and self.reader.buffer)
        if has_read_some and ts_not_changed and no_new_results:
            return True
        else:
            self._prev_datapoint_ts = self._last_datapoint_ts

    def get_resource_files(self):
        self._read_options()
        files = []
        if self.data_file:
            files.append(self.data_file)
        if self.errors_file:
            files.append(self.errors_file)
        return super(ExternalResultsLoader, self).get_resource_files() + files



"""
Module holds all stuff regarding Locust tool usage

Copyright 2015 BlazeMeter Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import json
import math
import os
import sys
from collections import OrderedDict, Counter

from bzt import TaurusConfigError
from bzt.engine import ScenarioExecutor, Scenario
from bzt.modules.aggregator import ConsolidatingAggregator, ResultsProvider, DataPoint, KPISet
from bzt.modules.console import ExecutorWidget
from bzt.modules.jmeter import JTLReader
from bzt.modules.services import PythonTool
from bzt.requests_model import HTTPRequest
from bzt.utils import iteritems, get_full_path, ensure_is_dict, PythonGenerator, FileReader, CALL_PROBLEMS
from bzt.utils import shutdown_process, dehumanize_time, RESOURCES_DIR


class LocustIOExecutor(ScenarioExecutor):
    def __init__(self):
        super(LocustIOExecutor, self).__init__()
        self.process = None
        self.is_master = False
        self.expected_workers = 0
        self.scenario = None
        self.script = None
        self.log_file = None
        self.locust = None

    def prepare(self):
        super(LocustIOExecutor, self).prepare()
        self.stdout = open(self.engine.create_artifact("locust", ".out"), 'w')
        self.stderr = open(self.engine.create_artifact("locust", ".err"), 'w')

        self.install_required_tools()
        self.scenario = self.get_scenario()
        self.__setup_script()
        self.engine.existing_artifact(self.script)

        # path to taurus dir. It's necessary for bzt usage inside tools/helpers
        self.env.add_path({"PYTHONPATH": get_full_path(__file__, step_up=3)})

        self.is_master = self.execution.get("master", self.is_master)

        if self.is_master:
            count_error = TaurusConfigError("Workers count required when starting in master mode")
            self.expected_workers = int(self.execution.get("workers", count_error))
            workers_ldjson = self.engine.create_artifact("locust-workers", ".ldjson")
            self.reader = WorkersReader(workers_ldjson, self.expected_workers, self.log)
            self.env.set({"WORKERS_LDJSON": workers_ldjson})
        else:
            kpi_jtl = self.engine.create_artifact("kpi", ".jtl")
            self.reader = JTLReader(kpi_jtl, self.log)
            self.env.set({"JTL": kpi_jtl})

        if isinstance(self.engine.aggregator, ConsolidatingAggregator):
            self.engine.aggregator.add_underling(self.reader)

    def install_required_tools(self):
        self.locust = self._get_tool(Locust, engine=self.engine, settings=self.settings)
        if not self.locust.check_if_installed():
            self.locust.install()

    def startup(self):
        load = self.get_load()
        concurrency = load.concurrency or 1
        run_time = 10  # seconds

        if self.is_master:
            concurrency = math.ceil(concurrency / float(self.expected_workers))

        if load.ramp_up:
            hatch = concurrency / float(load.ramp_up)
        else:
            hatch = concurrency

        wrapper = os.path.join(RESOURCES_DIR, "locustio-taurus-wrapper.py")

        if load.duration:
            run_time = dehumanize_time(load.duration)
            self.env.set({"LOCUST_DURATION": run_time})

        self.env.add_path({"PYTHONPATH": get_full_path(__file__, step_up=3)})
        self.env.add_path({"PYTHONPATH": self.engine.artifacts_dir})

        self.log_file = self.engine.create_artifact("locust", ".log")
        args = [sys.executable, wrapper, '-f', self.script]
        args += ['--logfile=%s' % self.log_file]
        args += ["--headless", "--only-summary", ]
        args += ["--users=%d" % concurrency, "--spawn-rate=%f" % hatch]
        args += ["--run-time=%d" % run_time, "--stop-timeout=10"]

        if load.iterations:
            num_requests = load.iterations * concurrency
            self.env.set({"LOCUST_NUMREQUESTS": num_requests})

        if self.is_master:
            args.extend(["--master", '--expect-workers=%s' % self.expected_workers])

        host = self.get_scenario().get("default-address")
        if host:
            args.append('--host=%s' % host)

        self.process = self._execute(args)

    def get_widget(self):
        """
        Add progress widget to console screen sidebar

        :rtype: ExecutorWidget
        """
        if not self.widget:
            label = "%s" % self
            self.widget = ExecutorWidget(self, "Locust.io: " + label.split('/')[1])
        return self.widget

    def check(self):
        retcode = self.process.poll()
        if retcode is not None:
            if retcode != 0:
                self.log.warning("Locust exited with non-zero code: %s", retcode)

            return True
        return False

    def resource_files(self):
        script = self.get_script_path()
        if script:
            return [script]
        else:
            return []

    def __tests_from_requests(self):
        filename = self.engine.create_artifact("generated_locust", ".py")
        locust_test = LocustIOScriptBuilder(self.scenario)
        locust_test.build_source_code()
        locust_test.save(filename)
        return filename

    def __setup_script(self):
        self.script = self.get_script_path()
        if not self.script:
            if "requests" in self.scenario:
                self.script = self.__tests_from_requests()
            else:
                msg = "There must be a script file or requests for its generation "
                msg += "to run Locust (%s)" % self.execution.get('scenario')
                raise TaurusConfigError(msg)

    def shutdown(self):
        shutdown_process(self.process, self.log)

    def post_process(self):
        self.locust.post_process()
        super(LocustIOExecutor, self).post_process()

    def has_results(self):
        master_results = self.is_master and self.reader.cumulative
        local_results = not self.is_master and self.reader and self.reader.buffer
        if master_results or local_results:
            return True
        else:
            return False

    def get_error_diagnostics(self):
        diagnostics = []
        if self.stdout is not None:
            with open(self.stdout.name) as fds:
                contents = fds.read().strip()
                if contents.strip():
                    diagnostics.append("Locust STDOUT:\n" + contents)
        if self.stderr is not None:
            with open(self.stderr.name) as fds:
                contents = fds.read().strip()
                if contents.strip():
                    diagnostics.append("Locust STDERR:\n" + contents)
        if self.log_file is not None and os.path.exists(self.log_file):
            with open(self.log_file) as fds:
                contents = fds.read().strip()
                if contents.strip():
                    diagnostics.append("Locust log:\n" + contents)
        return diagnostics


class Locust(PythonTool):
    PACKAGES = ["locust"]


class WorkersReader(ResultsProvider):
    def __init__(self, filename, num_workers, parent_logger):
        """
        :type filename: str
        :type num_workers: int
        :type parent_logger: logging.Logger
        """
        super(WorkersReader, self).__init__()
        self.log = parent_logger.getChild(self.__class__.__name__)
        self.join_buffer = {}
        self.num_workers = num_workers
        self.file = FileReader(filename=filename, parent_logger=self.log)
        self.read_buffer = ""

    def _calculate_datapoints(self, final_pass=False):
        read = self.file.get_bytes(size=1024 * 1024, last_pass=final_pass)
        if not read or not read.strip():
            return
        self.read_buffer += read
        while "\n" in self.read_buffer:
            _line = self.read_buffer[:self.read_buffer.index("\n") + 1]
            self.read_buffer = self.read_buffer[len(_line):]
            self.fill_join_buffer(json.loads(_line))

        max_full_ts = self.get_max_full_ts()

        if max_full_ts is not None:
            for point in self.merge_datapoints(max_full_ts):
                yield point

    def merge_datapoints(self, max_full_ts):
        reader_id = self.file.name + "@" + str(id(self))
        for key in sorted(self.join_buffer.keys(), key=int):
            if int(key) <= max_full_ts:
                sec_data = self.join_buffer.pop(key)
                self.log.debug("Processing complete second: %s", key)
                point = DataPoint(int(key))
                point[DataPoint.SOURCE_ID] = reader_id
                for sid, item in iteritems(sec_data):
                    point.merge_point(self.point_from_locust(key, sid, item))
                point.recalculate()
                yield point

    def get_max_full_ts(self):
        max_full_ts = None
        for key in sorted(self.join_buffer.keys(), key=int):
            if len(key) >= self.num_workers:
                max_full_ts = int(key)
        return max_full_ts

    def fill_join_buffer(self, data):
        self.log.debug("Got worker data: %s", data)
        for stats_item in data['stats']:
            for timestamp in stats_item['num_reqs_per_sec'].keys():
                if timestamp not in self.join_buffer:
                    self.join_buffer[timestamp] = {}
                self.join_buffer[timestamp][data['client_id']] = data

    @staticmethod
    def point_from_locust(timestamp, sid, data):
        """
        :type timestamp: str
        :type sid: str
        :type data: dict
        :rtype: DataPoint
        """
        point = DataPoint(int(timestamp))
        point[DataPoint.SOURCE_ID] = sid
        overall = KPISet()
        for item in data['stats']:
            if timestamp not in item['num_reqs_per_sec']:
                continue

            kpiset = KPISet()
            kpiset[KPISet.SAMPLE_COUNT] = item['num_reqs_per_sec'][timestamp]
            kpiset[KPISet.CONCURRENCY] = data['user_count']
            kpiset[KPISet.BYTE_COUNT] = item['total_content_length']
            if item['num_requests']:
                avg_rt = (item['total_response_time'] / 1000.0) / item['num_requests']
                kpiset.sum_rt = item['num_reqs_per_sec'][timestamp] * avg_rt

            for err in data['errors'].values():
                if err['name'] == item['name']:
                    new_err = KPISet.error_item_skel(err['error'], None, err['occurrences'], KPISet.ERRTYPE_ERROR,
                                                     Counter(), None)
                    KPISet.inc_list(kpiset[KPISet.ERRORS], ("msg", err['error']), new_err)
                    kpiset[KPISet.FAILURES] += err['occurrences']

            kpiset[KPISet.SUCCESSES] = kpiset[KPISet.SAMPLE_COUNT] - kpiset[KPISet.FAILURES]
            point[DataPoint.CURRENT][item['name']] = kpiset
            overall.merge_kpis(kpiset, sid)

        point[DataPoint.CURRENT][''] = overall
        point.recalculate()
        return point


class LocustIOScriptBuilder(PythonGenerator):
    IMPORTS = """
from gevent import sleep
from re import findall, compile
from locust import HttpUser, TaskSet, task, constant
"""

    def build_source_code(self):
        self.log.debug("Generating Python script for LocustIO")
        header_comment = self.gen_comment("This script was generated by Taurus", indent=0)
        scenario_class = self.gen_class_definition("UserBehaviour", ["TaskSet"])
        swarm_class = self.gen_class_definition("GeneratedSwarm", ["HttpUser"])
        imports = self.add_imports()

        self.root.append(header_comment)
        self.root.append(imports)
        self.root.append(scenario_class)
        self.root.append(swarm_class)

        swarm_class.append(self.gen_statement('tasks = [UserBehaviour]', indent=self.INDENT_STEP))

        default_address = self.scenario.get("default-address", "")
        swarm_class.append(self.gen_statement('host = "%s"' % default_address, indent=self.INDENT_STEP))

        swarm_class.append(self.gen_statement('wait_time = constant(%s)' % 0, indent=self.INDENT_STEP))
        swarm_class.append(self.gen_new_line())

        scenario_class.append(self.gen_decorator_statement('task(1)'))

        scenario_class.append(self.__gen_task())
        scenario_class.append(self.gen_new_line())

    def __gen_task(self):
        task = self.gen_method_definition("generated_task", ['self'])

        think_time = dehumanize_time(self.scenario.get_think_time())
        global_headers = self.scenario.get_headers()
        if not self.scenario.get("keepalive", True):
            global_headers['Connection'] = 'close'

        for req in self.scenario.get_requests():
            if not isinstance(req, HTTPRequest):
                msg = "Locust script generator doesn't support '%s' blocks, skipping"
                self.log.warning(msg, req.NAME)
                continue

            method = req.method.lower()
            if method not in ('get', 'delete', 'head', 'options', 'path', 'put', 'post'):
                raise TaurusConfigError("Wrong Locust request type: %s" % method)

            timeout = req.priority_option('timeout', default='30s')

            self.__gen_check(method, req, task, dehumanize_time(timeout), global_headers)

            if req.get_think_time():
                task.append(self.gen_statement("sleep(%s)" % dehumanize_time(req.get_think_time())))
            else:
                if think_time:
                    task.append(self.gen_statement("sleep(%s)" % think_time))
            task.append(self.gen_new_line())
        return task

    @staticmethod
    def __get_params_line(req, timeout, headers):
        param_dict = {'url': '"%s"' % req.url, 'timeout': timeout}
        if req.body:
            if isinstance(req.body, dict):
                param_dict['data'] = json.dumps(req.body)
            else:
                param_dict['data'] = '"%s"' % req.body

        if headers:
            param_dict['headers'] = json.dumps(headers)
        keys = (list(param_dict.keys()))
        keys.sort()
        return ', '.join(['%s=%s' % (key, param_dict[key]) for key in keys])

    def __gen_check(self, method, req, task, timeout, global_headers):
        assertions = req.config.get("assert", [])
        first_assert = True
        if assertions:
            statement = 'with self.client.%s(%s, catch_response=True) as response:'
        else:
            statement = "self.client.%s(%s)"
        headers = OrderedDict()
        if global_headers:
            sorted_headers = OrderedDict(sorted(global_headers.items(), key=lambda t: t[0]))
            headers.update(sorted_headers)
        if req.headers:
            headers.update(req.headers)
        task.append(self.gen_statement(statement % (method, self.__get_params_line(req, timeout, headers))))

        for idx, assertion in enumerate(assertions):
            assertion = ensure_is_dict(assertions, idx, "contains")
            if not isinstance(assertion['contains'], list):
                assertion['contains'] = [assertion['contains']]

            self.__gen_assertion(task, assertion, first_assert)
            first_assert = False

        if assertions:
            task.append(self.gen_statement('else:', indent=12))
            task.append(self.gen_statement('response.success()', indent=16))

    def __gen_assertion(self, task, assertion, is_first):
        subject = assertion.get("subject", Scenario.FIELD_BODY)
        values = [str(_assert) for _assert in assertion['contains']]
        if subject == 'body':
            content = 'response.content'
        elif subject == 'http-code':
            content = 'str(response.status_code)'
        else:
            raise TaurusConfigError('Wrong subject for Locust assertion: %s' % subject)

        if assertion.get('not', False):
            attr_not = ''
            func_name = 'any'
        else:
            attr_not = ' not'
            func_name = 'all'

        if assertion.get("regexp", True):
            expression = 'findall(compile(str(val)), %(content)s)' % {'content': content}
        else:
            expression = 'val in %s' % content

        statement = 'if%(not)s %(func)s(%(expression)s for val in %(values)s):'
        if subject == 'body':
            bin_values = [bytes(val, 'UTF-8') for val in values]
            statement = statement % {'not': attr_not, 'func': func_name, 'expression': expression, 'values': bin_values}
        else:
            statement = statement % {'not': attr_not, 'func': func_name, 'expression': expression, 'values': values}
        if not is_first:
            statement = 'el' + statement
        task.append(self.gen_statement(statement, indent=12))

        statement = 'response.failure("%(values)s%(not)s found in %(subject)s")'
        statement = statement % {'values': values, 'not': attr_not, 'subject': subject}
        task.append(self.gen_statement(statement, indent=16))
